from __future__ import division, print_function, absolute_import
import unittest
import math
import random

import numpy as num
import logging

from pyrocko import orthodrome, util
from pyrocko import orthodrome_ext
from .common import Benchmark
from pyrocko import config

logger = logging.getLogger('pyrocko.test.test_orthodrome')
benchmark = Benchmark()

earth_oblateness = 1./298.257223563
earthradius_equator = 6378.14 * 1000.
earthradius = config.config().earthradius

r2d = 180./math.pi
d2r = 1./r2d
km = 1000.

plot = False


def random_lat(mi=-90., ma=90., rstate=None, size=None):
    if rstate is None:
        rstate = num.random
    mi_ = 0.5*(math.sin(mi * math.pi/180.)+1.)
    ma_ = 0.5*(math.sin(ma * math.pi/180.)+1.)
    return num.arcsin(rstate.uniform(mi_, ma_, size=size)*2.-1.)*180./math.pi


def random_lon(mi=-180., ma=180., rstate=None, size=None):
    if rstate is None:
        rstate = num.random
    mi_ = 0.5*(math.sin(mi * math.pi/180.)+1.)
    ma_ = 0.5*(math.sin(ma * math.pi/180.)+1.)
    return num.arcsin(rstate.uniform(mi_, ma_, size=size)*2.-1.)*180./math.pi


def light(color, factor=0.5):
    return tuple(1-(1-c)*factor for c in color)


def dark(color, factor=0.5):
    return tuple(c*factor for c in color)


def have_geographiclib():
    try:
        from geographiclib.geodesic import Geodesic  # noqa
        return True

    except ImportError:
        return False


class OrthodromeTestCase(unittest.TestCase):

    @classmethod
    def tearDownClass(self):
        print(benchmark)

    def get_critical_random_locations(self, ntest):
        '''
        Create list of random (lat1, lon1, lat2, lon2) including critical
        locations.
        '''

        nasty_locations = [
            (0., 0., 0., 0.),
            (90., 0., -90., 0.),
            (0., -180., 0., 180.),
        ]

        assert ntest >= len(nasty_locations)

        lats1 = num.random.uniform(-90., 90., ntest)
        lats2 = num.random.uniform(-90., 90., ntest)
        lons1 = num.random.uniform(-180., 180., ntest)
        lons2 = num.random.uniform(-180., 180., ntest)

        for i, llll in enumerate(nasty_locations):
            lats1[i], lons1[i], lats2[i], lons2[i] = llll

        return lats1, lons1, lats2, lons2

    def testRaisesValueError(self):
        for lat1, lon1, lat2, lon2 in [
                (91., 0., 0., 0.),
                (0., 181., 0., 0.),
                (0., 0., 91., 0.),
                (0., 0., 0., 181.), ]:
            for f in [1, -1]:

                lat1 *= f
                lon1 *= f
                lat2 *= f
                lon2 *= f

                with self.assertRaises(ValueError):
                    orthodrome_ext.azibazi(lat1, lon1, lat2, lon2)

                with self.assertRaises(ValueError):
                    orthodrome_ext.distance_accurate50m(lat1, lon1, lat2, lon2)

    @benchmark
    def testAziBaziPython(self):
        ntest = 10000
        lats1, lons1, lats2, lons2 = self.get_critical_random_locations(ntest)
        for i in range(ntest):
            orthodrome.azibazi(
                float(lats1[i]), float(lons1[i]),
                float(lats2[i]), float(lons2[i]),
                implementation='python')

    @benchmark
    def testAziBaziC(self):
        ntest = 10000
        lats1, lons1, lats2, lons2 = self.get_critical_random_locations(ntest)
        for i in range(ntest):
            orthodrome.azibazi(
                lats1[i], lons1[i], lats2[i], lons2[i],
                implementation='c')

    def testAziBaziPythonC(self):
        ntest = 100
        lats1, lons1, lats2, lons2 = self.get_critical_random_locations(ntest)

        for i in range(ntest):
            azibazi_py = orthodrome.azibazi(
                float(lats1[i]), float(lons1[i]),
                float(lats2[i]), float(lons2[i]),
                implementation='python')
            azibazi_c = orthodrome.azibazi(
                lats1[i], lons1[i], lats2[i], lons2[i],
                implementation='c')

            num.testing.assert_almost_equal(azibazi_py, azibazi_c)

    @benchmark
    def testAziBaziArrayPython(self):
        ntest = 10000
        locs = self.get_critical_random_locations(ntest)
        orthodrome.azibazi_numpy(
            *locs,
            implementation='python')

    @benchmark
    def testAziBaziArrayC(self):
        ntest = 10000
        locs = self.get_critical_random_locations(ntest)
        orthodrome.azibazi_numpy(
            *locs,
            implementation='c')

    def testAziBaziArrayPythonC(self):
        ntest = 10000

        lats1, lons1, lats2, lons2 = self.get_critical_random_locations(ntest)

        azis_c, bazis_c = orthodrome.azibazi_numpy(
            lats1, lons1, lats2, lons2,
            implementation='c')

        azis_py, bazis_py = orthodrome.azibazi_numpy(
            lats1, lons1, lats2, lons2,
            implementation='python')

        for i in range(ntest):
            azi_py, bazi_py = orthodrome.azibazi(
                float(lats1[i]), float(lons1[i]),
                float(lats2[i]), float(lons2[i]),
                implementation='python')

            azi_c, bazi_c = orthodrome.azibazi(
                lats1[i], lons1[i], lats2[i], lons2[i],
                implementation='c')

            num.testing.assert_almost_equal(azi_py, azis_py[i])
            num.testing.assert_almost_equal(bazi_py, bazis_py[i])
            num.testing.assert_almost_equal(azi_c, azis_c[i])
            num.testing.assert_almost_equal(bazi_c, bazis_c[i])

        num.testing.assert_almost_equal(azis_py, azis_c)
        num.testing.assert_almost_equal(bazis_py, bazis_c)

    @benchmark
    def testDistancePython(self):
        ntest = 1000
        lats1, lons1, lats2, lons2 = self.get_critical_random_locations(ntest)
        loc1 = orthodrome.Loc(0., 0.)
        loc2 = orthodrome.Loc(0., 0.)
        for i in range(ntest):
            loc1.lat, loc1.lon = lats1[i], lons1[i]
            loc2.lat, loc2.lon = lats2[i], lons2[i]
            orthodrome.distance_accurate50m(
                loc1, loc2,
                implementation='python')

    @benchmark
    def testDistanceC(self):
        ntest = 1000
        lats1, lons1, lats2, lons2 = self.get_critical_random_locations(ntest)
        for i in range(ntest):
            orthodrome.distance_accurate50m(
                lats1[i], lons1[i], lats2[i], lons2[i],
                implementation='c')

    def testDistancePythonC(self):
        ntest = 100
        lats1, lons1, lats2, lons2 = self.get_critical_random_locations(ntest)
        for i in range(ntest):
            dist_py = orthodrome.distance_accurate50m(
                lats1[i], lons1[i], lats2[i], lons2[i],
                implementation='python')
            dist_c = orthodrome.distance_accurate50m(
                lats1[i], lons1[i], lats2[i], lons2[i],
                implementation='c')
            num.testing.assert_almost_equal(dist_py, dist_c)

    @benchmark
    def testDistanceArrayPython(self):
        ntest = 10000
        locs = self.get_critical_random_locations(ntest)
        orthodrome.distance_accurate50m_numpy(*locs, implementation='python')

    @benchmark
    def testDistanceArrayC(self):
        ntest = 10000
        locs = self.get_critical_random_locations(ntest)
        orthodrome.distance_accurate50m_numpy(*locs, implementation='c')

    def testDistanceArrayPythonC(self):
        ntest = 10000
        locs = self.get_critical_random_locations(ntest)
        a = orthodrome.distance_accurate50m_numpy(
            *locs,
            implementation='python')
        b = orthodrome.distance_accurate50m_numpy(
            *locs,
            implementation='c')
        num.testing.assert_array_almost_equal(a, b)

    def testGridDistances(self):
        for i in range(100):
            gsize = random.uniform(0., 1.)*2.*10.**random.uniform(4., 7.)
            north_grid, east_grid = num.meshgrid(
                num.linspace(-gsize/2., gsize/2., 11),
                num.linspace(-gsize/2., gsize/2., 11))

            north_grid = north_grid.flatten()
            east_grid = east_grid.flatten()

            lon = random.uniform(-180., 180.)
            lat = random.uniform(-90., 90.)

            lat_grid, lon_grid = orthodrome.ne_to_latlon(
                lat, lon, north_grid, east_grid)
            lat_grid_alt, lon_grid_alt = \
                orthodrome.ne_to_latlon_alternative_method(
                    lat, lon, north_grid, east_grid)

            for la, lo, no, ea in zip(
                    lat_grid, lon_grid, north_grid, east_grid):
                a = orthodrome.Loc(lat=la, lon=lo)
                b = orthodrome.Loc(lat=lat, lon=lon)

                cd = orthodrome.cosdelta(a, b)
                assert cd <= 1.0
                d = num.arccos(cd)*earthradius
                d2 = math.sqrt(no**2+ea**2)
                assert not (abs(d-d2) > 1.0e-3 and d2 > 1.)

    def test_wrap(self):
        assert orthodrome.wrap(11, -10, 10) == -9
        assert orthodrome.wrap(10, -10, 10) == -10
        assert orthodrome.wrap(10.001, -10, 10) == -9.999

    @unittest.skip('needs inspection')
    def test_local_distances(self):

        for reflat, reflon in [
                (0.0, 0.0),
                (10.0, 10.0),
                (90.0, 0.0),
                (-90.0, 0.0),
                (0.0, 180.0),
                (0.0, -180.0),
                (90.0, 180.0)]:

            north, east = serialgrid(num.linspace(-10*km, 10*km, 21),
                                     num.linspace(-10*km, 10*km, 21))

            lat, lon = orthodrome.ne_to_latlon2(reflat, reflon, north, east)
            north2, east2 = orthodrome.latlon_to_ne2(reflat, reflon, lat, lon)
            dist1 = num.sqrt(north**2 + east**2)
            dist2 = num.sqrt(north2**2 + east2**2)
            dist3 = orthodrome.distance_accurate15nm(reflat, reflon, lat, lon)
            assert num.all(num.abs(dist1-dist2) < 0.0001)
            assert num.all(num.abs(dist1-dist3) < 0.0001)

    def test_midpoint(self):
        center_lons = num.linspace(0., 180., 5)
        center_lats = [0., 89.]
        npoints = 10000
        half_side_length = 1000000.
        distance_error_max = 50000.
        for lat in center_lats:
            for lon in center_lons:
                n = num.random.uniform(
                    -half_side_length, half_side_length, npoints)
                e = num.random.uniform(
                    -half_side_length, half_side_length, npoints)
                dlats, dlons = orthodrome.ne_to_latlon(lat, lon, n, e)
                clat, clon = orthodrome.geographic_midpoint(dlats, dlons)
                d = orthodrome.distance_accurate50m_numpy(
                    clat, clon, lat, lon)[0]

                if plot:
                    import matplotlib.pyplot as plt
                    fig = plt.figure()
                    ax = fig.add_subplot(111)
                    ax.scatter(n, e)
                    c_n, c_e = orthodrome.latlon_to_ne_numpy(
                        lat, lon, clat, clon)
                    ax.plot(c_n, c_e, 'ro')
                    plt.show()

                self.assertTrue(d < distance_error_max, 'Distance %s > %s' %
                                (d, distance_error_max) +
                                '(maximum error)\n tested lat/lon: %s/%s' %
                                (lat, lon))

    @unittest.skipUnless(have_geographiclib(), 'geographiclib not available')
    def test_geodetic_to_ecef(self):
        orthodrome.geodetic_to_ecef(23., 0., 0.)

        a = orthodrome.earthradius_equator
        b = orthodrome.earthradius_equator * (1. - orthodrome.earth_oblateness)

        points = [
            ((90., 0., 0.), (0., 0., b)),
            ((-90., 0., 10.), (0., 0., -b-10.)),
            ((0., 0., 0.), (a, 0., 0.))]

        for p in points:
            num.testing.assert_almost_equal(orthodrome.geodetic_to_ecef(*p[0]),
                                            p[1])

    @unittest.skipUnless(have_geographiclib(), 'geographiclib not available')
    def test_ecef_to_geodetic(self):
        ncoords = 5
        lats = num.random.uniform(-90., 90, size=ncoords)
        lons = num.random.uniform(-180., 180, size=ncoords)
        alts = num.random.uniform(0, 10, size=ncoords)

        coords = num.array([lats, lons, alts]).T

        for ic in range(coords.shape[0]):
            xyz = orthodrome.geodetic_to_ecef(*coords[ic, :])
            latlonalt = orthodrome.ecef_to_geodetic(*xyz)

            num.testing.assert_almost_equal(coords[ic, :], latlonalt)

    def test_rotations(self):
        for lat in num.linspace(-90., 90., 20):
            for lon in num.linspace(-180., 180., 20):
                point = num.array([lat, lon], dtype=num.float)
                xyz = orthodrome.latlon_to_xyz(point)
                rot = orthodrome.rot_to_00(point[0], point[1])
                p2 = num.dot(rot, xyz)
                num.testing.assert_allclose(p2, [1., 0., 0.], atol=1.0e-7)

    def test_rotation2(self):
        eps = 1.0e-7
        lats = num.linspace(50., 60., 20)
        lons = num.linspace(170., 180., 20)
        lats2 = num.repeat(lats, lons.size)
        lons2 = num.tile(lons, lats.size)
        points = num.vstack((lats2, lons2)).T
        xyz = orthodrome.latlon_to_xyz(points)
        rot = orthodrome.rot_to_00(lats[0], lons[0])
        xyz2 = num.dot(rot, xyz.T).T
        points2 = orthodrome.xyz_to_latlon(xyz2)
        assert num.all(points2[:, 1] > -eps)

    def test_point_in_polygon(self):
        if plot:
            from pyrocko.plot import mpl_graph_color

            import matplotlib.pyplot as plt
            from matplotlib.patches import Polygon

            axes = plt.gca()

        nip = 100

        for i in range(1):
            np = 3
            points = num.zeros((np, 2))
            points[:, 0] = random_lat(size=3)
            points[:, 1] = random_lon(size=3)

            points_ip = num.zeros((nip*points.shape[0], 2))
            for ip in range(points.shape[0]):
                n, e = orthodrome.latlon_to_ne_numpy(
                    points[ip % np, 0], points[ip % np, 1],
                    points[(ip+1) % np, 0], points[(ip+1) % np, 1])

                ns = num.arange(nip) * n / nip
                es = num.arange(nip) * e / nip
                lats, lons = orthodrome.ne_to_latlon(
                    points[ip % np, 0], points[ip % np, 1], ns, es)

                points_ip[ip*nip:(ip+1)*nip, 0] = lats
                points_ip[ip*nip:(ip+1)*nip, 1] = lons

            if plot:
                color = mpl_graph_color(i)
                axes.add_patch(
                    Polygon(
                        num.fliplr(points_ip),
                        facecolor=light(color),
                        edgecolor=color,
                        alpha=0.5))

            points_xyz = orthodrome.latlon_to_xyz(points_ip.T)
            center_xyz = num.mean(points_xyz, axis=0)

            assert num.all(
                orthodrome.distances3d(
                    points_xyz, center_xyz[num.newaxis, :]) < 1.0)

            lat, lon = orthodrome.xyz_to_latlon(center_xyz)
            rot = orthodrome.rot_to_00(lat, lon)

            points_rot_xyz = num.dot(rot, points_xyz.T).T
            points_rot_pro = orthodrome.stereographic(points_rot_xyz)  # noqa

            poly_xyz = orthodrome.latlon_to_xyz(points_ip)
            poly_rot_xyz = num.dot(rot, poly_xyz.T).T
            groups = orthodrome.spoly_cut([poly_rot_xyz], axis=0)
            num.zeros(points.shape[0], dtype=num.int)

            if plot:
                for group in groups:
                    for poly_rot_group_xyz in group:

                        axes.set_xlim(-180., 180.)
                        axes.set_ylim(-90., 90.)

                    plt.show()

    def test_point_in_region(self):
        testdata = [
            ((-20., 180.), (-180., 180., -90., 90.), True),
            ((-20., 180.), (170., -170., -90., 90.), True),
            ((-20., 160.), (170., -170., -90., 90.), False),
            ((-20., -160.), (170., -170., -90., 90.), False),
        ]

        for point, region, in_region in testdata:
            assert bool(orthodrome.point_in_region(point, region)) == in_region


def serialgrid(x, y):
    return num.repeat(x, y.size), num.tile(y, x.size)


def plot_erroneous_ne_to_latlon():
    import gmtpy
    import random
    import subprocess
    import time

    while True:
        w, h = 20, 15

        gsize = random.uniform(0., 1.)*4.*10.**random.uniform(4., 7.)
        north_grid, east_grid = num.meshgrid(
            num.linspace(-gsize/2., gsize/2., 11),
            num.linspace(-gsize/2., gsize/2., 11))

        north_grid = north_grid.flatten()
        east_grid = east_grid.flatten()

        lat_delta = gsize/earthradius*r2d*2.
        lon = random.uniform(-180., 180.)
        lat = random.uniform(-90., 90.)

        print(gsize/1000.)

        lat_grid, lon_grid = orthodrome.ne_to_latlon(
            lat, lon, north_grid, east_grid)
        lat_grid_alt, lon_grid_alt = \
            orthodrome.ne_to_latlon_alternative_method(
                lat, lon, north_grid, east_grid)

        maxerrlat = num.max(num.abs(lat_grid-lat_grid_alt))
        maxerrlon = num.max(num.abs(lon_grid-lon_grid_alt))
        eps = 1.0e-8
        if maxerrlon > eps or maxerrlat > eps:
            print(lat, lon, maxerrlat, maxerrlon)

            gmt = gmtpy.GMT(
                config={
                    'PLOT_DEGREE_FORMAT': 'ddd.xxxF',
                    'PAPER_MEDIA': 'Custom_%ix%i' % (w*gmtpy.cm, h*gmtpy.cm),
                    'GRID_PEN_PRIMARY': 'thinnest/0/50/0'})

            south = max(-85., lat - 0.5*lat_delta)
            north = min(85., lat + 0.5*lat_delta)

            lon_delta = lat_delta/math.cos(lat*d2r)

            delta = lat_delta/360.*earthradius*2.*math.pi
            scale_km = gmtpy.nice_value(delta/10.)/1000.

            west = lon - 0.5*lon_delta
            east = lon + 0.5*lon_delta

            x, y = (west, east), (south, north)
            xax = gmtpy.Ax(mode='min-max', approx_ticks=4.)
            yax = gmtpy.Ax(mode='min-max', approx_ticks=4.)
            scaler = gmtpy.ScaleGuru(data_tuples=[(x, y)], axes=(xax, yax))
            scaler['R'] = '-Rg'
            layout = gmt.default_layout()
            mw = 2.5*gmtpy.cm
            layout.set_fixed_margins(
                mw, mw, mw/gmtpy.golden_ratio, mw/gmtpy.golden_ratio)
            widget = layout.get_widget()
            # widget['J'] =  ('-JT%g/%g'  % (lon, lat)) + '/%(width)gp'
            widget['J'] = (
                '-JE%g/%g/%g' % (lon, lat, min(lat_delta/2., 180.)))\
                + '/%(width)gp'
            aspect = gmtpy.aspect_for_projection(*(widget.J() + scaler.R()))
            widget.set_aspect(aspect)

            gmt.psbasemap(
                B='5g5',
                L=('x%gp/%gp/%g/%g/%gk' % (
                    widget.width()/2., widget.height()/7.,
                    lon, lat, scale_km)),
                *(widget.JXY()+scaler.R()))

            gmt.psxy(
                in_columns=(lon_grid, lat_grid),
                S='x10p', W='1p/200/0/0', *(widget.JXY()+scaler.R()))
            gmt.psxy(
                in_columns=(lon_grid_alt, lat_grid_alt),
                S='c10p', W='1p/0/0/200', *(widget.JXY()+scaler.R()))

            gmt.save('orthodrome.pdf')
            subprocess.call(['xpdf', '-remote', 'ortho', '-reload'])
            time.sleep(2)
        else:
            print('ok', gsize, lat, lon)


if __name__ == "__main__":
    plot = False
    util.setup_logging('test_orthodrome', 'warning')
    unittest.main(exit=False)
    print(benchmark)



from __future__ import division, print_function, absolute_import
import unittest
import math
import tempfile
import shutil

from pyrocko import model, util, trace, orthodrome, guts, moment_tensor
from pyrocko.guts import load
import numpy as num
from os.path import join as pjoin

d2r = num.pi/180.

eps = 1e-15


def assertOrtho(a, b, c):
    xeps = max(abs(max(a)), abs(max(b)), abs(max(c)))*eps
    assert abs(num.dot(a, b)) < xeps
    assert abs(num.dot(a, c)) < xeps
    assert abs(num.dot(b, c)) < xeps


def near(a, b, eps):
    return abs(a-b) < eps


class ModelTestCase(unittest.TestCase):

    def testIOEventOld(self):
        tempdir = tempfile.mkdtemp(prefix='pyrocko-model')
        fn = pjoin(tempdir, 'event.txt')
        e1 = model.Event(
            10., 20., 1234567890., 'bubu', region='taka tuka land',
            magnitude=5.1, magnitude_type='Mw',
            tags=['cluster:-1', 'custom_magnitude:2.5'])

        e1.olddump(fn)
        e2 = model.Event(load=fn)
        assert e1.region == e2.region
        assert e1.name == e2.name
        assert e1.lat == e2.lat
        assert e1.lon == e2.lon
        assert e1.time == e2.time
        assert e1.region == e2.region
        assert e1.magnitude == e2.magnitude
        assert e1.magnitude_type == e2.magnitude_type
        assert e1.tags == e2.tags
        shutil.rmtree(tempdir)

    def testIOEvent(self):
        tempdir = tempfile.mkdtemp(prefix='pyrocko-model')
        fn = pjoin(tempdir, 'event.txt')
        e1 = model.Event(
            10., 20., 1234567890., 'bubu', depth=10., region='taka tuka land',
            moment_tensor=moment_tensor.MomentTensor(strike=45., dip=90),
            magnitude=5.1, magnitude_type='Mw',
            tags=['cluster:-1', 'custom_magnitude:2.5'])

        guts.dump(e1, filename=fn)
        e2 = guts.load(filename=fn)
        assert e1.region == e2.region
        assert e1.name == e2.name
        assert e1.lat == e2.lat
        assert e1.lon == e2.lon
        assert e1.time == e2.time
        assert e1.region == e2.region
        assert e1.magnitude == e2.magnitude
        assert e1.magnitude_type == e2.magnitude_type
        assert e1.get_hash() == e2.get_hash()
        assert e1.tags == e2.tags

        fn2 = pjoin(tempdir, 'events.txt')
        guts.dump_all([e1, e2], filename=fn2)

        with self.assertRaises(model.OneEventRequired):
            model.load_one_event(fn2)

        shutil.rmtree(tempdir)

    def testMissingComponents(self):

        ne = model.Channel('NE', azimuth=45., dip=0.)
        se = model.Channel('SE', azimuth=135., dip=0.)

        station = model.Station('', 'STA', '', 0., 0., 0., channels=[ne, se])

        mat = station.projection_to_enu(('NE', 'SE', 'Z'), ('E', 'N', 'U'))[0]
        assertOrtho(mat[:, 0], mat[:, 1], mat[:, 2])

        n = model.Channel('D', azimuth=0., dip=90.)
        station.set_channels([n])
        mat = station.projection_to_enu(('N', 'E', 'D'), ('E', 'N', 'U'))[0]
        assertOrtho(mat[:, 0], mat[:, 1], mat[:, 2])

    def testIOStations(self):
        tempdir = tempfile.mkdtemp(prefix='pyrocko-model')
        fn = pjoin(tempdir, 'stations.txt')

        ne = model.Channel('NE', azimuth=45., dip=0.)
        se = model.Channel('SE', azimuth=135., dip=0.)
        stations = [
            model.Station('', sta, '', 0., 0., 0., channels=[ne, se])
            for sta in ['STA1', 'STA2']]

        model.dump_stations(stations, fn)
        stations = model.load_stations(fn)

        shutil.rmtree(tempdir)

    def testProjections(self):
        km = 1000.

        ev = model.Event(lat=-10, lon=150., depth=0.0)

        for azi in num.linspace(0., 360., 37):
            lat, lon = orthodrome.ne_to_latlon(
                ev.lat, ev.lon, 10.*km * math.cos(azi), 10.*km * math.sin(azi))

            sta = model.Station(lat=lat, lon=lon)
            sta.set_event_relative_data(ev)
            sta.set_channels_by_name('BHZ', 'BHE', 'BHN')

            r = 1.
            t = 1.
            traces = [
                trace.Trace(
                    channel='BHE',
                    ydata=num.array([math.sin(azi)*r+math.cos(azi)*t])),
                trace.Trace(
                    channel='BHN',
                    ydata=num.array([math.cos(azi)*r-math.sin(azi)*t])),
            ]

            for m, in_channels, out_channels in sta.guess_projections_to_rtu():
                projected = trace.project(traces, m, in_channels, out_channels)

            def g(traces, cha):
                for tr in traces:
                    if tr.channel == cha:
                        return tr

            r = g(projected, 'R')
            t = g(projected, 'T')
            assert(near(r.ydata[0], 1.0, 0.001))
            assert(near(t.ydata[0], 1.0, 0.001))

    def testProjectionsZOnly(self):
        km = 1000.

        ev = model.Event(lat=-10, lon=150., depth=0.0)

        for azi in num.linspace(0., 360., 37):
            lat, lon = orthodrome.ne_to_latlon(
                ev.lat, ev.lon, 10.*km * math.cos(azi), 10.*km * math.sin(azi))

            sta = model.Station(lat=lat, lon=lon)
            sta.set_event_relative_data(ev)
            sta.set_channels_by_name('BHZ', 'BHN', 'BHE')
            traces = [
                trace.Trace(
                    channel='BHZ',
                    ydata=num.array([1.0])),
            ]

            projected = []
            for m, in_channels, out_channels in sta.guess_projections_to_enu():
                projected.extend(trace.project(
                    traces, m, in_channels, out_channels))

            def g(traces, cha):
                for tr in traces:
                    if tr.channel == cha:
                        return tr

            z = g(projected, 'U')
            assert(near(z.ydata[0], 1.0, 0.001))

    def getGNSSCampaign(self, nstations=25, nsparse=False):
        rstate = num.random.RandomState(None)

        lats = rstate.uniform(90, -90, nstations)
        lons = rstate.uniform(90, -90, nstations)

        shifts = rstate.uniform(-2.5, 2.5, (nstations, 3))
        sigma = rstate.uniform(-0.5, 0.5, (nstations, 3))

        correlations = rstate.uniform(-0.5, 0.5, (nstations, 3))

        campaign = model.gnss.GNSSCampaign()

        for ista in range(nstations):

            north = model.gnss.GNSSComponent(
                shift=float(shifts[ista, 0]),
                sigma=float(sigma[ista, 0]))
            east = model.gnss.GNSSComponent(
                shift=float(shifts[ista, 1]),
                sigma=float(sigma[ista, 1]))
            up = model.gnss.GNSSComponent(
                shift=float(shifts[ista, 2]),
                sigma=float(sigma[ista, 2]))

            station = model.gnss.GNSSStation(
                lat=float(lats[ista]),
                lon=float(lons[ista]),
                correlation_eu=float(correlations[ista, 0]),
                correlation_ne=float(correlations[ista, 1]),
                correlation_nu=float(correlations[ista, 2]),
                north=north,
                east=east,
                up=up)

            campaign.add_station(station)

        if nsparse:
            assert nsparse < nstations, 'nsparse < nstations'

            stations2 = list(campaign.stations)
            rstate.shuffle(stations2)
            stations = stations2[:nsparse]
            channels = rstate.randint(0, 3, size=nsparse)

            channel_map = {
                0: 'north',
                1: 'east',
                2: 'up'
            }

            for ista, sta in enumerate(stations):
                cha = channels[ista]
                sta.__setattr__(channel_map[cha], None)

        return campaign

    def testGNSSCampaign(self):
        tempdir = tempfile.mkdtemp(prefix='pyrocko-model')
        fn = pjoin(tempdir, 'gnss_campaign.yml')

        campaign = self.getGNSSCampaign()

        campaign.dump(filename=fn)
        campaign2 = load(filename=fn)

        s1 = campaign.stations[0]

        s_add = s1.north + s1.north
        assert s_add.shift == (s1.north.shift + s1.north.shift)

        assert len(campaign.stations) == len(campaign2.stations)
        shutil.rmtree(tempdir)

    def testGNSSCampaignCorrelationMatrix(self):
        campaign = self.getGNSSCampaign()

        corr_arr = campaign.get_correlation_matrix()
        num.testing.assert_array_equal(corr_arr, corr_arr.T)

        corr_arr_ref = num.zeros(corr_arr.shape)

        idx = 0
        for ista, sta in enumerate(campaign.stations):
            ncomp = sta.ncomponents
            components = sta.components.values()

            for ic, comp in enumerate(components):
                corr_arr_ref[idx+ic, idx+ic] = comp.sigma

            corr_arr_ref[idx, idx+1] = sta.correlation_ne
            corr_arr_ref[idx+1, idx] = sta.correlation_ne

            corr_arr_ref[idx, idx+2] = sta.correlation_nu
            corr_arr_ref[idx+2, idx] = sta.correlation_nu

            corr_arr_ref[idx+1, idx+2] = sta.correlation_eu
            corr_arr_ref[idx+2, idx+1] = sta.correlation_eu

            idx += ncomp

        num.testing.assert_array_equal(corr_arr, corr_arr_ref)

    def testGNSSCampaignCovarianceMatrix(self):
        campaign = self.getGNSSCampaign(2)

        covar_arr = campaign.get_covariance_matrix()
        num.testing.assert_array_equal(covar_arr, covar_arr.T)

        covar_ref = num.zeros(covar_arr.shape)

        idx = 0
        for ista, sta in enumerate(campaign.stations):
            ncomp = sta.ncomponents

            for ic, comp in enumerate([sta.north, sta.east, sta.up]):
                covar_ref[ista*3+ic, ista*3+ic] = comp.sigma**2

            covar_ref[ista*3, ista*3+1] =\
                sta.correlation_ne * sta.north.sigma * sta.east.sigma
            covar_ref[ista*3+1, ista*3] =\
                sta.correlation_ne * sta.north.sigma * sta.east.sigma

            covar_ref[ista*3, ista*3+2] =\
                sta.correlation_nu * sta.north.sigma * sta.up.sigma
            covar_ref[ista*3+2, ista*3] =\
                sta.correlation_nu * sta.north.sigma * sta.up.sigma

            covar_ref[ista*3+1, ista*3+2] =\
                sta.correlation_eu * sta.east.sigma * sta.up.sigma
            covar_ref[ista*3+2, ista*3+1] =\
                sta.correlation_eu * sta.east.sigma * sta.up.sigma

            idx += ncomp

        num.testing.assert_array_equal(covar_arr, covar_ref)

    def testGNSSCampaignSparse(self):
        nstations = 20
        nsparse = 5
        campaign = self.getGNSSCampaign(nstations=nstations, nsparse=nsparse)

        cor_arr = campaign.get_correlation_matrix()
        size = cor_arr.shape[0]
        assert size == nstations*3 - nsparse
        num.testing.assert_array_equal(cor_arr, cor_arr.T)

        cov_arr = campaign.get_covariance_matrix()
        size = cov_arr.shape[0]
        assert size == nstations*3 - nsparse
        num.testing.assert_array_equal(cov_arr, cov_arr.T)

        comp_mask = num.concatenate(
            [s.get_component_mask() for s in campaign.stations])
        assert comp_mask.sum() == campaign.ncomponents


if __name__ == "__main__":
    util.setup_logging('test_trace', 'warning')
    unittest.main()



from __future__ import division, print_function, absolute_import
import unittest
from . import common

import pyrocko.trace
from pyrocko import util, io, model, pile

if common.have_obspy():
    import obspy
    from pyrocko import obspy_compat
    obspy_compat.plant()


def close_win(win):
    win.close()


@common.require_obspy
class ObsPyCompatTestCase(unittest.TestCase):

    @common.require_gui
    def test_obspy_snuffle(self):
        fn = common.test_data_file('test1.mseed')

        stream = obspy.read(fn)
        stream.snuffle(launch_hook=close_win)

        trace = stream[0]
        trace.snuffle(launch_hook=close_win)

    @common.require_gui
    def test_obspy_fiddle(self):
        fn = common.test_data_file('test1.mseed')

        stream = obspy.read(fn)
        stream2 = stream.fiddle(launch_hook=close_win)  # noqa

        trace = stream[0]
        trace2 = trace.fiddle(launch_hook=close_win)  # noqa

    def test_to_obspy_trace(self):
        traces = io.load(common.test_data_file('test1.mseed'))
        for tr in traces:
            obs_tr = tr.to_obspy_trace()

            assert isinstance(obs_tr, obspy.Trace)
            assert obs_tr.data.size == tr.data_len()

            obs_stats = obs_tr.stats
            for attr in ('network', 'station', 'location', 'channel'):
                assert obs_stats.__getattr__(attr) == tr.__getattribute__(attr)

    def test_to_obspy_stream(self):
        pl = pile.Pile()
        pl.load_files([common.test_data_file('test1.mseed')],
                      show_progress=False)
        st = pl.to_obspy_stream()

        assert isinstance(st, obspy.Stream)
        assert len(st) == len([tr for tr in pl.iter_all()])
        for tr in st:
            assert isinstance(tr, obspy.Trace)

    def test_to_pyrocko_traces(self):
        st = obspy.read(common.test_data_file('test1.mseed'))

        traces = st.to_pyrocko_traces()
        assert isinstance(traces, list)
        for tr in traces:
            assert isinstance(tr, pyrocko.trace.Trace)

        for tr in st:
            assert isinstance(tr.to_pyrocko_trace(), pyrocko.trace.Trace)

    def test_to_pyrocko_stations(self):
        fn = common.test_data_file('geeil.geofon.xml')
        inventory = obspy.read_inventory(fn)

        for sta in inventory.to_pyrocko_stations():
            assert isinstance(sta, model.Station)

    def test_to_pyrocko_events(self):
        from obspy.clients.fdsn.client import Client
        client = Client('IRIS')
        cat = client.get_events(eventid=609301)
        events = cat.to_pyrocko_events()
        self.assertEqual(len(events), len(cat))


if __name__ == "__main__":
    util.setup_logging('test_obspy_compat', 'warning')
    unittest.main()



# http://pyrocko.org - GPLv3
#
# The Pyrocko Developers, 21st Century
# ---|P------/S----------~Lg----------
from __future__ import absolute_import, division

try:
    from urllib.request import Request
    from urllib.error import HTTPError
    from urllib.parse import urlencode
    from future.moves.urllib.request import urlopen
except ImportError:
    from urllib2 import Request, urlopen
    from urllib2 import HTTPError
    from urllib import urlencode
import calendar
import logging

from pyrocko import util, model
from pyrocko.moment_tensor import MomentTensor
from .base_catalog import EarthquakeCatalog, NotFound

logger = logging.getLogger('pyrocko.client.kinherd')

km = 1000.


def ws_request(url, post=False, **kwargs):
    url_values = urlencode(kwargs)
    url = url + '?' + url_values
    logger.debug('Accessing URL %s' % url)

    req = Request(url)
    if post:
        req.add_data(post)

    req.add_header('Accept', '*/*')

    try:
        return urlopen(req)

    except HTTPError as e:
        if e.code == 404:
            raise NotFound(url)
        else:
            raise e


class Kinherd(EarthquakeCatalog):

    def __init__(self, catalog='KPS'):
        self.catalog = catalog
        self.events = {}

    def retrieve(self, **kwargs):
        import yaml

        kwargs['format'] = 'yaml'

        url = 'http://kinherd.org/quakes/%s' % self.catalog

        f = ws_request(url, **kwargs)

        names = []
        for eq in yaml.safe_load_all(f):
            pset = eq['parametersets'][0]
            tref = calendar.timegm(pset['reference_time'].timetuple())
            tpost = calendar.timegm(pset['posted_time'].timetuple())
            params = pset['parameters']

            mt = MomentTensor(
                strike=params['strike'],
                dip=params['dip'],
                rake=params['slip_rake'],
                scalar_moment=params['moment'])

            event = model.Event(
                time=tref + params['time'],
                lat=params['latitude'],
                lon=params['longitude'],
                depth=params['depth'],
                magnitude=params['magnitude'],
                duration=params['rise_time'],
                name=eq['name'],
                catalog=self.catalog,
                moment_tensor=mt)

            event.ext_confidence_intervals = {}
            trans = {'latitude': 'lat', 'longitude': 'lon'}
            for par in 'latitude longitude depth magnitude'.split():
                event.ext_confidence_intervals[trans.get(par, par)] = \
                    (params[par+'_ci_low'], params[par+'_ci_high'])

            event.ext_posted_time = tpost

            name = eq['name']
            self.events[name] = event
            names.append(name)

        return names

    def iter_event_names(self, time_range=None, **kwargs):

        qkwargs = {}
        for k in 'magmin magmax latmin latmax lonmin lonmax'.split():
            if k in kwargs and kwargs[k] is not None:
                qkwargs[k] = '%f' % kwargs[k]

        if time_range is not None:
            form = '%Y-%m-%d_%H-%M-%S'
            if time_range[0] is not None:
                qkwargs['tmin'] = util.time_to_str(time_range[0], form)
            if time_range[1] is not None:
                qkwargs['tmax'] = util.time_to_str(time_range[1], form)

        for name in self.retrieve(**qkwargs):
            yield name

    def get_event(self, name):
        if name not in self.events:
            self.retrieve(name=name)

        return self.events[name]



from __future__ import division, print_function, absolute_import

import sys
import os
import logging
import unittest
import numpy as num

from pyrocko import util
from pyrocko import eikonal_ext

from . import common

assert_ae = num.testing.assert_almost_equal


logger = logging.getLogger('pyrocko.test.test_eikonal')

benchmark = common.Benchmark()

km = 1000.


show_plot = int(os.environ.get('MPL_SHOW', 0))

def is_64bits():
    return sys.maxsize > 2**32


class EikonalTestCase(unittest.TestCase):

    @classmethod
    def tearDownClass(cls):
        print(benchmark)

    def fpath_reference(self, fn):
        try:
            return common.test_data_file(fn)
        except util.DownloadError:
            return common.test_data_file_no_download(fn)

    def compare_with_reference(self, arr, fn, rtol=1e-9):
        fpath = self.fpath_reference(fn)
        if not os.path.exists(fpath):
            num.save(fpath, arr)

        arr_ref = num.load(fpath)
        finite_ref = num.isfinite(arr_ref)
        finite = num.isfinite(arr)

        assert num.allclose(arr_ref[finite_ref], arr[finite], rtol=rtol)

    def test_empty(self):

        speeds = num.zeros((0, 0))
        times = num.zeros((0, 0))
        delta = 1.0

        with self.assertRaises(eikonal_ext.EikonalExtError):
            eikonal_ext.eikonal_solver_fmm_cartesian(speeds, times, delta)

    def test_nd(self):
        ndim_max = 3
        n = 51
        for ndim in range(1, ndim_max+2):
            shape = (n,) * ndim
            speeds = num.ones(shape)
            times = num.zeros(shape) - 1.0
            delta = 2.0 / (n-1)

            iseed = tuple(i//2 for i in shape)
            times[iseed] = 0.0

            if ndim <= ndim_max:
                @benchmark.labeled('test_nd: %iD' % ndim)
                def run():
                    eikonal_ext.eikonal_solver_fmm_cartesian(
                        speeds, times, delta)

                run()

                fshapes = []
                for idim in range(ndim):
                    fshape = [1] * ndim
                    fshape[idim] = n
                    fshapes.append(tuple(fshape))

                xs = [
                    num.linspace(-1.0, 1.0, n).reshape(fshape)
                    for fshape in fshapes]

                times_ref = num.sqrt(sum(x**2 for x in xs))
                assert num.max(num.abs(times-times_ref)) \
                    < 1e6 + (ndim-1) * delta

            else:
                with self.assertRaises(eikonal_ext.EikonalExtError):
                    eikonal_ext.eikonal_solver_fmm_cartesian(
                        speeds, times, delta)

    def test_earthmodel(self):
        from pyrocko import cake
        mod = cake.load_model()

        nx, ny = 500, 500
        delta = (cake.earthradius * 2.0) / (nx-1)

        x = num.arange(nx) * delta - cake.earthradius
        y = num.arange(ny) * delta - cake.earthradius

        x2 = x[num.newaxis, :]
        y2 = y[:, num.newaxis]

        z = cake.earthradius - num.sqrt(x2**2 + y2**2)

        vp_pro = mod.profile('vp')
        z_pro = mod.profile('z')

        vp = num.interp(z, z_pro, vp_pro)

        inside = z > 0.0

        speeds = num.ones((ny, nx))
        speeds[:, :] = 300.
        speeds[inside] = vp[inside]

        times = num.zeros((ny, nx)) - 1.0

        iy = ny - int(round(600*km / delta))
        ix = nx//2

        times[iy, ix] = 0.0

        @benchmark.labeled('test_earthmodel')
        def run():
            eikonal_ext.eikonal_solver_fmm_cartesian(speeds, times, delta)

        run()

        times[num.logical_not(inside)] = num.nan
        self.compare_with_reference(times, 'test_earthmodel.npy')

        if show_plot:
            from matplotlib import pyplot as plt

            plt.gcf().add_subplot(1, 1, 1, aspect=1.0)
            plt.pcolormesh(x, y, speeds, cmap='gray', edgecolor='none')
            plt.contour(x, y, times, levels=num.linspace(0., 1200, 20))
            plt.gca().axis('off')
            plt.show()

    def test_2d(self):

        nx, ny = 1000, 500
        delta = 20. / float(nx)
        x = num.arange(nx) * delta - 10.0
        y = num.arange(ny) * delta - 5.0

        x2 = x[num.newaxis, :]
        y2 = y[:, num.newaxis]
        speeds = num.ones((ny, nx))  # + num.sin(x2)*0.1
        r1 = num.sqrt((x2-3)**2 + y2**2)
        r2 = num.sqrt((x2+3)**2 + y2**2)

        speeds[r1 < 2.0] = 1.5
        speeds[r2 < 2.0] = 0.66

        rstate = num.random.RandomState(23)
        speeds *= rstate.uniform(0.5, 1.5, (ny, nx))
        times = num.zeros((ny, nx)) - 1.0
        # times[:, 0] = y**2 / 10.0
        # times[:, 0] = y*10
        times[0, :] = (x-num.min(x)) * 0.2
        # times[0, 0] = 0.0
        # times[0, -1] = 0.0
        # times[ny//2, nx//2] = 0.0

        @benchmark.labeled('test_2d')
        def run():
            eikonal_ext.eikonal_solver_fmm_cartesian(speeds, times, delta)

        run()

        self.compare_with_reference(
            times, 'test_2d.npy',
            rtol=(1e-9 if is_64bits() else 1e-2))  # should be investigated

        if show_plot:
            from matplotlib import pyplot as plt

            plt.gcf().add_subplot(1, 1, 1, aspect=1.0)
            plt.contourf(x, y, times)
            plt.contourf(x, y, speeds, alpha=0.1, cmap='gray', )
            plt.show()

    def test_2d_snail(self):

        nx, ny = 1000, 500
        delta = 20. / float(nx)
        x = num.arange(nx) * delta - 10.0
        y = num.arange(ny) * delta - 5.0

        # x2 = x[num.newaxis, :]
        # y2 = y[:, num.newaxis]
        speeds = num.ones((ny, nx))  # + num.sin(x2)*0.1

        a = num.linspace(0.5, 4.0, 6000)

        sx = a * num.sin(a*(2.*num.pi))
        sy = a * num.cos(a*(2.*num.pi))

        isx = nx//2 + num.round(sx / delta).astype(num.int)
        isy = ny//2 + num.round(sy / delta).astype(num.int)

        speeds[isy, isx] = 0.00001

        times = num.zeros((ny, nx)) - 1.0
        times[ny//2, 0] = 0.0

        @benchmark.labeled('test_2d_snail')
        def run():
            eikonal_ext.eikonal_solver_fmm_cartesian(speeds, times, delta)

        run()

        self.compare_with_reference(times, 'test_2d_snail.npy')

        if show_plot:
            from matplotlib import pyplot as plt

            plt.gcf().add_subplot(1, 1, 1, aspect=1.0)
            times[times > 50.] = num.nan
            plt.contourf(
                x, y, times,
                levels=num.linspace(0., 50., 50))
            plt.contour(
                x, y, times,
                levels=num.linspace(0., 50., 50),
                colors='white',
                alpha=0.5)
            plt.contourf(
                x, y, speeds,
                alpha=0.1, cmap='gray')

            plt.show()

    def test_3d(self):

        nx, ny, nz = 100, 100, 100
        delta = 10. / float(nx)
        x = num.arange(nx) * delta - 5.0
        y = num.arange(ny) * delta - 5.0
        z = num.arange(nz) * delta - 5.0

        # x2 = x[num.newaxis, num.newaxis, :]
        # y2 = y[num.newaxis, :, num.newaxis]
        # z2 = z[:, num.newaxis, num.newaxis]

        speeds = num.ones((nz, ny, nx))  # + num.sin(x2)*0.1

        times = num.zeros((ny, nx, nz)) - 1.0
        times[nz//2, ny//2, nx//2] = 0.0

        @benchmark.labeled('test_3d')
        def run():
            eikonal_ext.eikonal_solver_fmm_cartesian(speeds, times, delta)

        run()

        self.compare_with_reference(times, 'test_3d.npy')

        if show_plot:
            from mpl_toolkits.mplot3d import axes3d  # noqa
            from matplotlib import pyplot as plt

            x3, y3 = num.meshgrid(x, y)

            ax = plt.gca(projection='3d')
            ax.scatter(
                x3.flatten(), y3.flatten(), z[0], c=times[0, :, :].flatten())

            x3, z3 = num.meshgrid(x, z)

            ax = plt.gca(projection='3d')
            ax.scatter(
                x3.flatten(), y[0], z3.flatten(), c=times[:, 0, :].flatten())
            y3, z3 = num.meshgrid(y, z)

            ax = plt.gca(projection='3d')
            ax.scatter(
                x[0], y3.flatten(), z3.flatten(), c=times[:, :, 0].flatten())
            plt.show()

    def benchmark_against_fast_sweeping(self):
        from beat.fast_sweeping import fast_sweep

        nx, ny = 300, 300
        delta = 20. / float(nx)
        x = num.arange(nx) * delta
        y = num.arange(ny) * delta

        # x2 = x[num.newaxis, :]
        # y2 = y[:, num.newaxis]
        speeds = num.ones((ny, nx))  # + num.sin(x2)*0.1
        speeds[ny//2:, :] = 3.0

        slownesses = 1.0/speeds

        times_fm = num.zeros((ny, nx)) - 1.0
        times_fm[ny//4, nx//4] = 0.0

        @benchmark.labeled('fast marching')
        def fm():
            eikonal_ext.eikonal_solver_fmm_cartesian(speeds, times_fm, delta)

        times_fs = []

        @benchmark.labeled('fast sweeping')
        def fs():
            times_fs.append(fast_sweep.get_rupture_times_c(
                slownesses.flatten(), delta,
                nx, ny, ny//4, nx//4).reshape((ny, nx)))

        fm()
        fs()

        times_fs = times_fs[0]

        from matplotlib import pyplot as plt

        def get_elapsed(name):
            for xname, elapsed in benchmark.results:
                if name == xname:
                    return elapsed

        plt.gcf().add_subplot(1, 3, 1, aspect=1.0)
        plt.title('fast marching, max=%g, t_cpu=%.2g s' % (
            num.max(times_fm), get_elapsed('fast marching')))
        plt.contourf(x, y, times_fm)

        plt.gcf().add_subplot(1, 3, 2, aspect=1.0)
        plt.title('fast sweeping, max=%g, t_cpu=%.2g s' % (
            num.max(times_fs), get_elapsed('fast sweeping')))
        plt.contourf(x, y, times_fs)

        plt.gcf().add_subplot(1, 3, 3, aspect=1.0)
        plt.title('difference, |max|=%g' % num.max(num.abs(times_fm-times_fs)))
        plt.contourf(x, y, times_fm-times_fs)

        plt.show()


if __name__ == '__main__':
    util.setup_logging('test_eikonal', 'warning')
    unittest.main()



from __future__ import division, print_function, absolute_import
import math
import unittest
import os
import numpy as num

from pyrocko import gf, util

r2d = 180. / math.pi
d2r = 1.0 / r2d
km = 1000.
show_plot = int(os.environ.get('MPL_SHOW', 0))


def numeq(a, b, eps):
    return (num.all(num.asarray(a).shape == num.asarray(b).shape and
            num.abs(num.asarray(a) - num.asarray(b)) < eps))


class GFSTFTestCase(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        unittest.TestCase.__init__(self, *args, **kwargs)

    def test_stf_triangular(self):
        from matplotlib import pyplot as plt

        tref = 10.
        for duration in [0., 3.]:
            for peak_ratio in num.linspace(0., 1., 11):
                stf = gf.TriangularSTF(
                    duration=duration, peak_ratio=peak_ratio, anchor=0.)
                t, a = stf.discretize_t(deltat=0.1, tref=tref)
                assert numeq(stf.centroid_time(tref), tref, 1e-5)
                if show_plot:
                    plt.title('Triangular')
                    plt.plot(t, a)
                    plt.plot(t, a, 'o')

        if show_plot:
            plt.show()

    def test_stf_boxcar(self):
        from matplotlib import pyplot as plt

        for duration in [0., 1., 2., 3.]:
            tref = 10.02
            stf = gf.BoxcarSTF(duration=duration, anchor=0.)
            t, a = stf.discretize_t(deltat=0.1, tref=tref)
            assert numeq(stf.centroid_time(tref), tref, 1e-5)
            if show_plot:
                plt.title('Boxcar')
                plt.plot(t, a)
                plt.plot(t, a, 'o')

        if show_plot:
            plt.show()

    def test_stf_half_sinusoid(self):
        from matplotlib import pyplot as plt

        for duration in [0., 1., 2., 3.]:
            tref = 10.02
            stf = gf.HalfSinusoidSTF(duration=duration, anchor=0.)
            t, a = stf.discretize_t(deltat=0.1, tref=tref)
            assert numeq(stf.centroid_time(tref), tref, 1e-5)
            if show_plot:
                plt.title('Half Sinosoid')
                plt.plot(t, a)
                plt.plot(t, a, 'o')

        if show_plot:
            plt.show()

    def test_stf_resonator(self):
        from matplotlib import pyplot as plt

        duration = 30.
        frequency = 1./15
        tref = 20.
        deltat = 1.0

        stf = gf.ResonatorSTF(duration=duration, frequency=frequency)
        t, a = stf.discretize_t(deltat=0.1, tref=tref)
        if show_plot:
            plt.title('Resonator')
            plt.plot(t, a)
            plt.plot(t, a, 'o')

        if show_plot:
            plt.show()

    def test_effective_durations(self):
        deltat = 1e-4
        for stf in [
                gf.HalfSinusoidSTF(duration=2.0),
                gf.TriangularSTF(duration=2.0, peak_ratio=0.),
                gf.TriangularSTF(duration=2.0, peak_ratio=1.),
                gf.TriangularSTF(duration=2.0, peak_ratio=0.5),
                gf.BoxcarSTF(duration=2.0)]:

            t, a = stf.discretize_t(deltat, 0.0)
            t0 = stf.centroid_time(0.0)

            edur = num.sqrt(num.sum((t-t0)**2 * a)) * 2. * num.sqrt(3.)
            assert abs(edur - stf.effective_duration) < 1e-3

    def test_objects(self):
        for stf_class in gf.seismosizer.stf_classes:
            if stf_class in (gf.STF, gf.ResonatorSTF):
                continue

            d1 = 2.0
            stf = stf_class(effective_duration=d1)
            d2 = stf.effective_duration
            assert abs(d2 - d1) < 1e-4


if __name__ == '__main__':
    show_plot = True
    util.setup_logging('test_gf_stf', 'warning')
    unittest.main()



from __future__ import division, print_function, absolute_import
import unittest
import tempfile

from pyrocko import util, model
from pyrocko.gui import marker


class MarkerTestCase(unittest.TestCase):

    def test_writeread(self):
        nslc_ids = [('', 'STA', '', '*')]

        event = model.Event(lat=111., lon=111., depth=111., time=111.)

        _marker = marker.Marker(nslc_ids=nslc_ids, tmin=1., tmax=10.)
        emarker = marker.EventMarker(event=event)
        pmarker = marker.PhaseMarker(nslc_ids=nslc_ids,  tmin=1., tmax=10.)
        pmarker.set_event(event)

        emarker.set_alerted(True)

        markers = [_marker, emarker, pmarker]
        fn = tempfile.mkstemp()[1]

        marker.save_markers(markers, fn)

        in_markers = marker.load_markers(fn)
        in__marker, in_emarker, in_pmarker = in_markers
        for i, m in enumerate(in_markers):
            if not isinstance(m, marker.EventMarker):
                assert (m.tmax - m.tmin) == 9.
            else:
                assert not m.is_alerted()

        marker.associate_phases_to_events([in_pmarker, in_emarker])

        in_event = in_pmarker.get_event()

        assert all((in_event.lat == 111., in_event.lon == 111.,
                    in_event.depth == 111., in_event.time == 111.))

        assert event.get_hash() == in_event.get_hash()
        assert in_pmarker.get_event_hash() == in_event.get_hash()
        assert in_pmarker.get_event_time() == 111.


if __name__ == "__main__":
    util.setup_logging('test_marker', 'warning')
    unittest.main()



# http://pyrocko.org - GPLv3
#
# The Pyrocko Developers, 21st Century
# ---|P------/S----------~Lg----------
from __future__ import absolute_import
from future import standard_library
standard_library.install_aliases()  # noqa

import logging
import os
from collections import namedtuple
import os.path as op

import numpy as num

from pyrocko import util, config
from pyrocko import orthodrome as od

logger = logging.getLogger('pyrocko.dataset.geonames')

GeoName = namedtuple('GeoName', '''
geonameid
name
asciiname
alt_names
lat
lon
feature_class
feature_code
country_code
alt_country_code
admin1_code
admin2_code
admin3_code
admin4_code
population
elevation
dem
timezone
modification_date'''.split())

GeoName2 = namedtuple('GeoName2', '''
name
asciiname
lat
lon
population
feature_code
'''.split())

base_url = 'http://download.geonames.org/export/dump/'


def download_file(fn, dirpath):
    import urllib.request
    url = base_url + '/' + fn
    fpath = op.join(dirpath, fn)
    logger.info('starting download of %s' % url)

    util.ensuredirs(fpath)
    f = urllib.request.urlopen(url)
    fpath_tmp = fpath + '.%i.temp' % os.getpid()
    g = open(fpath_tmp, 'wb')
    while True:
        data = f.read(1024)
        if not data:
            break
        g.write(data)

    g.close()
    f.close()

    os.rename(fpath_tmp, fpath)

    logger.info('finished download of %s' % url)


def positive_region(region):
    west, east, south, north = [float(x) for x in region]

    assert -180. - 360. <= west < 180.
    assert -180. < east <= 180. + 360.
    assert -90. <= south < 90.
    assert -90. < north <= 90.

    if east < west:
        east += 360.

    if west < -180.:
        west += 360.
        east += 360.

    return (west, east, south, north)


def ascii_str(u):
    return u.encode('ascii', 'replace')


def load(zfn, fn, minpop=1000000, region=None):
    geonames_dir = config.config().geonames_dir
    filepath = op.join(geonames_dir, zfn or fn)
    if not os.path.exists(filepath):
        download_file(zfn or fn, geonames_dir)

    if region:
        w, e, s, n = positive_region(region)

    if zfn is not None:
        import zipfile
        z = zipfile.ZipFile(filepath, 'r')
        f = z.open(fn, 'r')
    else:
        z = None
        f = open(filepath, 'rb')

    for line in f:
        t = line.split(b'\t')
        pop = int(t[14])
        if minpop <= pop:
            lat = float(t[4])
            lon = float(t[5])
            feature_code = str(t[7].decode('ascii'))
            if not region or (
                    (w <= lon <= e or w <= lon + 360. <= e)
                    and (s <= lat <= n)):

                yield GeoName2(
                    t[1].decode('utf8'),
                    str(t[2].decode('utf8').encode('ascii', 'replace')
                        .decode('ascii')),
                    lat, lon, pop, feature_code)

    f.close()
    if z is not None:
        z.close()


g_cities = {}


def load_all_keep(zfn, fn, minpop=1000000, region=None, exclude=()):
    if (zfn, fn) not in g_cities:
        g_cities[zfn, fn] = list(load(zfn, fn, minpop=0))

    if region:
        w, e, s, n = positive_region(region)
        return [c for c in g_cities[zfn, fn]
                if (minpop <= c.population and
                    ((w <= c.lon <= e or w <= c.lon + 360. <= e)
                        and (s <= c.lat <= n)) and
                    c.feature_code not in exclude)]

    else:
        return [c for c in g_cities[zfn, fn] if (
            minpop <= c.population and c.feature_code not in exclude)]


def get_cities_region(region=None, minpop=0):
    return load_all_keep(
        'cities1000.zip', 'cities1000.txt',
        region=region,
        minpop=minpop,
        exclude=('PPLX',))


def get_cities_by_name(name):
    cities = get_cities_region()
    candidates = []
    for c in cities:
        if c.asciiname.lower() == name.lower():
            candidates.append(c)

    return candidates


def get_cities(lat, lon, radius, minpop=0):
    region = od.radius_to_region(lat, lon, radius)
    cities = get_cities_region(region, minpop=minpop)

    clats = num.array([c.lat for c in cities])
    clons = num.array([c.lon for c in cities])

    dists = od.distance_accurate50m_numpy(lat, lon, clats, clons)
    order = num.argsort(dists)
    cities_sorted = [cities[i] for i in order if dists[i] < radius]
    return cities_sorted



from __future__ import division, print_function, absolute_import

import unittest
import calendar
import math
import re
import sys
import datetime
from contextlib import contextmanager


from pyrocko.guts import StringPattern, Object, Bool, Int, Float, String, \
    SObject, Unicode, Complex, Timestamp, DateTimestamp, StringChoice, Defer, \
    ArgumentError, ValidationError, Any, List, Tuple, Union, Choice, Dict, \
    load, load_string, load_xml_string, load_xml, load_all, iload_all, \
    load_all_xml, iload_all_xml, dump, dump_xml, dump_all, dump_all_xml, \
    make_typed_list_class, walk, zip_walk, path_to_str, clone, set_elements, \
    get_elements, YPathError


guts_prefix = 'guts_test'


class SamplePat(StringPattern):
    pattern = r'[a-z]{3}'


class SampleChoice(StringChoice):
    choices = ['a', 'bcd', 'efg']


basic_types = (
    Bool, Int, Float, String, Unicode, Complex, Timestamp, SamplePat,
    SampleChoice)


def tstamp(*args):
    return float(calendar.timegm(args))


samples = {}
samples[Bool] = [True, False]
samples[Int] = [2**n for n in [1, 30]]  # ,31,65] ]
samples[Float] = [0., 1., math.pi, float('inf'), float('-inf'), float('nan')]
samples[String] = [
    '', 'test', 'abc def', '<', '\n', '"', '\'',
    ''.join(chr(x) for x in range(32, 128))]
# chr(0) and other special chars don't work with xml...

samples[Unicode] = [u'aoeu \u00e4 \u0100']
samples[Complex] = [1.0+5J, 0.0J, complex(math.pi, 1.0)]
samples[Timestamp] = [
    0.0,
    tstamp(2030, 1, 1, 0, 0, 0),
    tstamp(1960, 1, 1, 0, 0, 0),
    tstamp(2010, 10, 10, 10, 10, 10) + 0.000001]

samples[SamplePat] = ['aaa', 'zzz']
samples[SampleChoice] = ['a', 'bcd', 'efg']

regularize = {}
regularize[Bool] = [(1, True), (0, False), ('0', False), ('False', False)]
regularize[Int] = [('1', 1), (1.0, 1), (1.1, 1)]
regularize[Float] = [
    ('1.0', 1.0),
    (1, 1.0),
    ('inf', float('inf')),
    ('nan', float('nan'))]
regularize[String] = [(1, '1')]
regularize[Unicode] = [(1, u'1')]
regularize[Timestamp] = [
    ('2010-01-01 10:20:01',  tstamp(2010, 1, 1, 10, 20, 1)),
    ('2010-01-01T10:20:01',  tstamp(2010, 1, 1, 10, 20, 1)),
    ('2010-01-01T10:20:01.11Z',  tstamp(2010, 1, 1, 10, 20, 1)+0.11),
    ('2030-12-12 00:00:10.11111',  tstamp(2030, 12, 12, 0, 0, 10)+0.11111),
    (datetime.date(2010, 12, 12),  tstamp(2010, 12, 12, 0, 0, 0)),
    (10,  10.)
]
regularize[DateTimestamp] = [
    ('2010-01-01',  tstamp(2010, 1, 1, 0, 0, 0)),
    (datetime.datetime(2010, 12, 12),  tstamp(2010, 12, 12, 0, 0, 0)),
    (datetime.date(2010, 12, 12),  tstamp(2010, 12, 12, 0, 0, 0)),
    (86400, 86400.)
]

regularize[Complex] = [
    ([1., 2.], 1.0+2.0J),
    ((1., 2.), 1.0+2.0J)
]


class GutsTestCase(unittest.TestCase):

    if sys.version_info < (2, 7):

        @contextmanager
        def assertRaises(self, exc):

            gotit = False
            try:
                yield None
            except exc:
                gotit = True

            assert gotit, 'expected to get a %s exception' % exc

        def assertIsNone(self, value):
            assert value is None, 'expected None but got %s' % value

    def assertEqualNanAware(self, a, b):
        if isinstance(a, float) \
                and isinstance(b, float) \
                and math.isnan(a) and math.isnan(b):
            return

        if isinstance(a, float) and isinstance(b, float):
            self.assertEqual('%.14g' % a, '%.14g' % b)
        else:
            self.assertEqual(a, b)

    def testStringChoice(self):
        class X(Object):
            m = StringChoice.T(['a', 'b'])

        x = X(m='a')
        x.validate()
        x = X(m='c')
        with self.assertRaises(ValidationError):
            x.validate()

    def testStringChoiceIgnoreCase(self):
        class X(Object):
            m = StringChoice.T(['a', 'b'], ignore_case=True)

        x = X(m='A')
        x.validate()
        x = X(m='C')
        with self.assertRaises(ValidationError):
            x.validate()

    def testDefaultMaker(self):
        class X(Object):
            a = Int.T(default=1)

        class Y(Object):
            x = X.T(default=X.D(a=2))

        class Yb(Object):
            x = X.T(default=X(a=2))

        class Z(Object):
            y = Y.T(default=Y.D(x=X.D(a=3)))

        class Zb(Object):
            y = Y.T(default=Yb(x=X(a=3)))

        class Zl(Object):
            y = Y.T(default=Y.D(x=X.D(a=3)))

        class Zbl(Object):
            y = Y.T(default=Yb(x=X(a=3)))

        y = Y()
        self.assertEqual(y.x.a, 2)

        z = Z()
        z2 = Z()

        z2.y.x.a = 5
        self.assertEqual(z.y.x.a, 3)
        self.assertEqual(z2.y.x.a, 5)

        y = Yb()
        self.assertEqual(y.x.a, 2)

        z = Zb()
        z2 = Zb()

        z2.y.x.a = 5
        self.assertEqual(z.y.x.a, 3)
        self.assertEqual(z2.y.x.a, 5)

    def testDefaultList(self):
        class X(Object):
            a = Int.T(default=1)

        class Y(Object):
            xs = List.T(X.T(), default=[X(a=2)])

        y = Y()
        y.xs[0].a = 3
        y2 = Y()
        self.assertEqual(y2.xs[0].a, 2)
        self.assertEqual(y.xs[0].a, 3)

    def testDefaultTuple(self):
        class X(Object):
            a = Int.T(default=1)

        class Y(Object):
            xs = Tuple.T(1, X.T(), default=(X(a=2),))

        y = Y()
        y.xs[0].a = 3
        y2 = Y()
        self.assertEqual(y2.xs[0].a, 2)
        self.assertEqual(y.xs[0].a, 3)

    def testDefaultDict(self):
        class X(Object):
            a = Int.T(default=1)

        class Y(Object):
            xs = Dict.T(X.T(), X.T(), default={
                X(a=2): X(a=3)})

        def fi(d):
            return list(d.values())[0]

        y = Y()
        fi(y.xs).a = 4
        y2 = Y()
        self.assertEqual(fi(y2.xs).a, 3)
        self.assertEqual(fi(y.xs).a, 4)

    def testRemove(self):
        class X(Object):
            y = Int.T(default=1, xmlstyle='attribute')
            z = List.T(Int.T())
            a = Int.T(xmlstyle='content')

        x = X(a=11, z=[1, 2, 3])
        del x

        X.T.remove_property('y')
        X.T.remove_property('z')
        X.T.remove_property('a')

        assert len(X.T.propnames) == 0
        assert len(X.T.xmltagname_to_name) == 0
        assert len(X.T.xmltagname_to_name_multivalued) == 0
        assert len(X.T.xmltagname_to_name_multivalued) == 0

    def testAny(self):
        class X(Object):
            y = Int.T(default=1)

        class A(Object):
            x = Any.T()
            lst = List.T(Any.T())

        a1 = A(x=X(y=33))
        a1.validate()
        a1c = load_string(a1.dump())
        self.assertEqual(a1c.x.y, 33)

        a2 = A(x=22, lst=['a', 44, X(y=22)])
        a2.validate()
        a2c = load_string(a2.dump())
        self.assertEqual(a2c.x, 22)
        self.assertEqual(a2c.lst[:2], ['a', 44])
        self.assertEqual(a2c.lst[2].y, 22)

        a3 = A(x=X(y='10'))
        with self.assertRaises(ValidationError):
            a3.validate()

        a3.regularize()
        assert a3.x.y == 10

    def testAnyObject(self):
        class Base(Object):
            pass

        class A(Base):
            y = Int.T(default=1)

        class B(Base):
            x = Float.T(default=1.0)

        class C(Object):
            c = Base.T(optional=True)

        ca = C(c=A())
        cb = C(c=B())
        c_err = C(c='abc')
        cc_err = C(c=C())
        ca_err = C(c=A(y='1'))

        for c in (ca, cb):
            c.validate()

        for c in (c_err, cc_err, ca_err):
            with self.assertRaises(ValidationError):
                c.validate()

        ca_err.regularize()
        assert ca_err.c.y == 1

    def testUnion(self):

        class X1(Object):
            xmltagname = 'root'
            m = Union.T(members=[Int.T(), StringChoice.T(['small', 'large'])])

        class U(Union):
            members = [Int.T(), StringChoice.T(['small', 'large'])]

        class X2(Object):
            xmltagname = 'root'
            m = U.T()

        for X in [X1, X2]:
            X = X1
            x1 = X(m='1')
            with self.assertRaises(ValidationError):
                x1.validate()

            x1.validate(regularize=True)
            self.assertEqual(x1.m, 1)

            x2 = X(m='small')
            x2.validate()
            x3 = X(m='fail!')
            with self.assertRaises(ValidationError):
                x3.validate()
            with self.assertRaises(ValidationError):
                x3.validate(regularize=True)

            for x in [x1, x2]:
                y = load_string(x.dump())
                self.assertEqual(x.m, y.m)

            for x in [x1, x2]:
                y = load_xml_string(x.dump_xml())
                self.assertEqual(x.m, y.m)

    def testChoice(self):
        class A(Object):
            x = Int.T(default=1)

        class B(Object):
            y = Int.T(default=2)

        class OhOh(Object):
            z = Float.T(default=4.0)

        class AB(Choice):
            choices = [A.T(xmltagname='a'), B.T(xmltagname='b')]

        class C1(Object):
            xmltagname = 'root1'
            ab = Choice.T(choices=[A.T(xmltagname='a'), B.T(xmltagname='b')])

        class C2(Object):
            xmltagname = 'root2'
            ab = AB.T()

        for C in (C1, C2):
            for good in (A(), B()):
                c = C(ab=good)
                c.validate()
                for dumpx, loadx in (
                        (dump, load_string),
                        (dump_xml, load_xml_string)):
                    c2 = loadx(dumpx(c))
                    self.assertEqual(type(c2), C)
                    self.assertEqual(type(c2.ab), type(good))

            for bad in (OhOh(), 5.0):
                c = C(ab=bad)
                with self.assertRaises(ValidationError):
                    c.validate()

        c = C1(ab=A(x='5'))
        c.regularize()
        assert c.ab.x == 5
        c = C1(ab=B(y='6'))
        c.regularize()
        assert c.ab.y == 6

    def testRegularizeChoice(self):

        class A(Object):
            x = Choice.T(optional=True, choices=[Timestamp.T(), Int.T()])

        a = A()
        a.regularize()

        a = A(x='2015-01-01 00:00:00')
        a.regularize()

        a = A(x='5')
        a.regularize()

        a = A(x='abc')
        with self.assertRaises(ValidationError):
            a.regularize()

    def testTooMany(self):

        class A(Object):
            m = List.T(Int.T())
            xmltagname = 'a'

        a = A(m=[1, 2, 3])

        class A(Object):
            m = Int.T()
            xmltagname = 'a'

        with self.assertRaises(ArgumentError):
            load_xml_string(a.dump_xml())

    def testDeferred(self):
        class A(Object):
            p = Defer('B.T', optional=True)

        class B(Object):
            p = A.T()

        a = A(p=B(p=A()))
        del a

    def testListDeferred(self):
        class A(Object):
            p = List.T(Defer('B.T'))
            q = List.T(Defer('B.T'))

        class B(Object):
            p = List.T(A.T())

        a = A(p=[B(p=[A()])], q=[B(), B()])
        del a

    def testOptionalList(self):
        class A(Object):
            l1 = List.T(Int.T())
            l2 = List.T(Int.T(), optional=True)
            l3 = List.T(Int.T(), default=[1, 2, 3])

        a = A()
        a.validate()
        self.assertEqual(a.l1, [])
        self.assertIsNone(a.l2)
        self.assertEqual(a.l3, [1, 2, 3])

    def testSelfDeferred(self):
        class A(Object):
            a = Defer('A.T', optional=True)

        a = A(a=A(a=A()))
        del a

    def testOrder(self):
        class A(Object):
            a = Int.T(default=1)
            b = Int.T(default=2)

        class B(A):
            a = Int.T(default=3, position=A.a.position)
            c = Int.T(default=4)

        b = B()
        names = [k for (k, v) in b.T.inamevals(b)]
        assert names == ['a', 'b', 'c']

    def testOrderDeferred(self):
        class A(Object):
            a = Defer('B.T')
            b = Int.T(default=0)

        class B(Object):
            x = Int.T(default=1)

        a = A(a=B())
        names = [k for (k, v) in a.T.inamevals(a)]

        assert names == ['a', 'b']

    def testDocs(self):
        class A(Object):
            '''A description'''

            a = Int.T(default=0, help='the a property')

        class B(A):
            '''B description'''

            b = Float.T(default=0, help='the b property')
            c = List.T(Int.T(), help='the c')

        self.assertEqual(B.__doc__, '''B description

    .. py:gattribute:: b

      ``float``, *default:* ``0``
      
      the b property

    .. py:gattribute:: c

      ``list`` of ``int`` objects, *default:* ``[]``
      
      the c
''')  # noqa

    def testContentStyleXML(self):

        class Duration(Object):
            unit = String.T(optional=True, xmlstyle='attribute')
            uncertainty = Float.T(optional=True)
            value = Float.T(optional=True, xmlstyle='content')

            xmltagname = 'duration'

        s = '<duration unit="s"><uncertainty>1.0</uncertainty>10.5</duration>'
        dur = load_xml_string(s)
        self.assertEqual(dur.value, float('10.5'))
        self.assertEqual(dur.unit, 's')
        self.assertEqual(dur.uncertainty, float('1.0'))
        self.assertEqual(re.sub(r'\n\s*', '', dur.dump_xml()), s)

    def testSObject(self):

        class DotName(SObject):
            network = String.T()
            station = String.T()

            def __init__(self, s=None, **kwargs):
                if s is not None:
                    network, station = s.split('.')
                    kwargs = dict(network=network, station=station)

                SObject.__init__(self, **kwargs)

            def __str__(self):
                return '.'.join((self.network, self.station))

        class X(Object):
            xmltagname = 'root'
            dn = DotName.T()

        x = X(dn=DotName(network='abc', station='def'))
        self.assertEqual(x.dn.network, 'abc')
        self.assertEqual(x.dn.station, 'def')
        x = load_string(dump(x))
        self.assertEqual(x.dn.network, 'abc')
        self.assertEqual(x.dn.station, 'def')
        x = load_xml_string(dump_xml(x))
        self.assertEqual(x.dn.network, 'abc')
        self.assertEqual(x.dn.station, 'def')

    def testDict(self):
        class A(Object):
            d = Dict.T(Int.T(), Float.T())

        a = A(d={1: 2.0})

        with self.assertRaises(NotImplementedError):
            a.dump_xml()

        a2 = load_string(a.dump())
        self.assertEqual(a2.dump(), a.dump())

        for (k1, v1), (k2, v2) in [
                [('1', 2.0), (1, 2.0)],
                [(1, '2.0'), (1, 2.0)],
                [('1', '2.0'), (1, 2.0)]]:

            a = A(d={k1: v1})
            with self.assertRaises(ValidationError):
                a.validate()

            a.regularize()
            self.assertEqual(a.d[k2], v2)

    def testOptionalDefault(self):

        from pyrocko.guts_array import Array, array_equal
        import numpy as num
        assert_ae = num.testing.assert_almost_equal

        def array_equal_noneaware(a, b):
            if a is None:
                return b is None
            elif b is None:
                return a is None
            else:
                return array_equal(a, b)

        data = [
            ('a', Int.T(),
                [None, 0, 1, 2],
                ['aerr', 0, 1, 2]),
            ('b', Int.T(optional=True),
                [None, 0, 1, 2],
                [None, 0, 1, 2]),
            ('c', Int.T(default=1),
                [None, 0, 1, 2],
                [1, 0, 1, 2]),
            ('d', Int.T(default=1, optional=True),
                [None, 0, 1, 2],
                [1, 0, 1, 2]),
            ('e', List.T(Int.T()),
                [None, [], [1], [2]],
                [[], [], [1], [2]]),
            ('f', List.T(Int.T(), optional=True),
                [None, [], [1], [2]],
                [None, [], [1], [2]]),
            ('g', List.T(Int.T(), default=[1]), [
                None, [], [1], [2]],
                [[1], [], [1], [2]]),
            ('h', List.T(Int.T(), default=[1], optional=True),
                [None, [], [1], [2]],
                [[1], [], [1], [2]]),
            ('i', Tuple.T(2, Int.T()),
                [None, (1, 2)],
                ['err', (1, 2)]),
            ('j', Tuple.T(2, Int.T(), optional=True),
                [None, (1, 2)],
                [None, (1, 2)]),
            ('k', Tuple.T(2, Int.T(), default=(1, 2)),
                [None, (1, 2), (3, 4)],
                [(1, 2), (1, 2), (3, 4)]),
            ('l', Tuple.T(2, Int.T(), default=(1, 2), optional=True),
                [None, (1, 2), (3, 4)],
                [(1, 2), (1, 2), (3, 4)]),
            ('i2', Tuple.T(None, Int.T()),
                [None, (1, 2)],
                [(), (1, 2)]),
            ('j2', Tuple.T(None, Int.T(), optional=True),
                [None, (), (3, 4)],
                [None, (), (3, 4)]),
            ('k2', Tuple.T(None, Int.T(), default=(1,)),
                [None, (), (3, 4)],
                [(1,), (), (3, 4)]),
            ('l2', Tuple.T(None, Int.T(), default=(1,), optional=True),
                [None, (), (3, 4)],
                [(1,), (), (3, 4)]),
            ('m', Array.T(shape=(None,), dtype=num.int, serialize_as='list'),
                [num.arange(0), num.arange(2)],
                [num.arange(0), num.arange(2)]),
            ('n', Array.T(shape=(None,), dtype=num.int, serialize_as='list',
                          optional=True),
                [None, num.arange(0), num.arange(2)],
                [None, num.arange(0), num.arange(2)]),
            ('o', Array.T(shape=(None,), dtype=num.int, serialize_as='list',
                          default=num.arange(2)),
                [None, num.arange(0), num.arange(2), num.arange(3)],
                [num.arange(2), num.arange(0), num.arange(2), num.arange(3)]),
            ('p', Array.T(shape=(None,), dtype=num.int, serialize_as='list',
                          default=num.arange(2), optional=True),
                [None, num.arange(0), num.arange(2), num.arange(3)],
                [num.arange(2), num.arange(0), num.arange(2), num.arange(3)]),
            ('q', Dict.T(String.T(), Int.T()),
                [None, {}, {'a': 1}],
                [{}, {}, {'a': 1}]),
            ('r', Dict.T(String.T(), Int.T(), optional=True),
                [None, {}, {'a': 1}],
                [None, {}, {'a': 1}]),
            ('s', Dict.T(String.T(), Int.T(), default={'a': 1}),
                [None, {}, {'a': 1}],
                [{'a': 1}, {}, {'a': 1}]),
            ('t', Dict.T(String.T(), Int.T(), default={'a': 1}, optional=True),
                [None, {}, {'a': 1}],
                [{'a': 1}, {}, {'a': 1}]),
        ]

        for k, t, vals, exp, in data:
            last = [None]

            class A(Object):
                def __init__(self, **kwargs):
                    last[0] = len(kwargs)
                    Object.__init__(self, **kwargs)

                v = t

            A.T.class_signature()

            for v, e in zip(vals, exp):
                if isinstance(e, str) and e == 'aerr':
                    with self.assertRaises(ArgumentError):
                        if v is not None:
                            a1 = A(v=v)
                        else:
                            a1 = A()

                    continue
                else:
                    if v is not None:
                        a1 = A(v=v)
                    else:
                        a1 = A()

                if isinstance(e, str) and e == 'err':
                    with self.assertRaises(ValidationError):
                        a1.validate()
                else:
                    a1.validate()
                    a2 = load_string(dump(a1))
                    if isinstance(e, num.ndarray):
                        assert last[0] == int(
                            not (array_equal_noneaware(t.default(), a1.v)
                                 and t.optional))
                        assert_ae(a1.v, e)
                        assert_ae(a1.v, e)
                    else:
                        assert last[0] == int(
                            not(t.default() == a1.v and t.optional))
                        self.assertEqual(a1.v, e)
                        self.assertEqual(a2.v, e)

    def testArray(self):
        from pyrocko.guts_array import Array
        import numpy as num

        shapes = [(None,), (1,), (10,), (1000,)]
        for shape in shapes:
            for serialize_as in ('base64', 'table', 'npy', 'base64+meta'):
                class A(Object):
                    xmltagname = 'aroot'
                    arr = Array.T(
                        shape=shape,
                        dtype=num.int,
                        serialize_as=serialize_as,
                        serialize_dtype='>i8')

                n = shape[0] or 10
                a = A(arr=num.arange(n, dtype=num.int))
                b = load_string(a.dump())
                self.assertTrue(num.all(a.arr == b.arr))

                if serialize_as != 'base64+meta':
                    b = load_xml_string(a.dump_xml())
                    self.assertTrue(num.all(a.arr == b.arr))

                if shape[0] is not None:
                    with self.assertRaises(ValidationError):
                        a = A(arr=num.arange(n+10, dtype=num.int))
                        a.validate()

        for s0 in [None, 2, 10]:
            for s1 in [None, 2, 10]:
                class A(Object):
                    xmltagname = 'aroot'
                    arr = Array.T(
                        shape=(s0, s1),
                        dtype=num.int)

                n0 = s0 or 100
                n1 = s1 or 100
                a = A(arr=num.arange(n0*n1, dtype=num.int).reshape((n0, n1)))
                b = load_string(a.dump())
                self.assertTrue(num.all(a.arr == b.arr))

                b = load_xml_string(a.dump_xml())
                self.assertTrue(num.all(a.arr == b.arr))

    def testArrayNoDtype(self):
        from pyrocko.guts_array import Array
        import numpy as num

        class A(Object):
            arr = Array.T(serialize_as='base64+meta')

        for dtype in (num.int, num.float):
            a = A(arr=num.zeros((3, 3), dtype=dtype))
            b = load_string(a.dump())
            assert a.arr.shape == b.arr.shape
            self.assertTrue(num.all(a.arr == b.arr))

        b_int = load_string  # noqa

    def testPO(self):
        class SKU(StringPattern):
            pattern = '\\d{3}-[A-Z]{2}'

        class Comment(String):
            xmltagname = 'comment'

        class Quantity(Int):
            pass

        class USAddress(Object):
            country = String.T(
                default='US', optional=True, xmlstyle='attribute')
            name = String.T()
            street = String.T()
            city = String.T()
            state = String.T()
            zip = Float.T()

        class Item(Object):
            part_num = SKU.T(xmlstyle='attribute')
            product_name = String.T()
            quantity = Quantity.T()
            us_price = Float.T(xmltagname='USPrice')
            comment = Comment.T(optional=True)
            ship_date = DateTimestamp.T(optional=True)

        class Items(Object):
            item_list = List.T(Item.T())

        class PurchaseOrderType(Object):
            order_date = DateTimestamp.T(optional=True, xmlstyle='attribute')
            ship_to = USAddress.T()
            bill_to = USAddress.T()
            comment = Comment.T(optional=True)
            items = Items.T()

        class PurchaseOrder(PurchaseOrderType):
            xmltagname = 'purchaseOrder'

        xml = '''<?xml version="1.0"?>
<purchaseOrder orderDate="1999-10-20">
   <shipTo country="US">
      <name>Alice Smith</name>
      <street>123 Maple Street</street>
      <city>Mill Valley</city>
      <state>CA</state>
      <zip>90952</zip>
   </shipTo>
   <billTo country="US">
      <name>Robert Smith</name>
      <street>8 Oak Avenue</street>
      <city>Old Town</city>
      <state>PA</state>
      <zip>95819</zip>
   </billTo>
   <comment>Hurry, my lawn is going wild</comment>
   <items>
      <item partNum="872-AA">
         <productName>Lawnmower</productName>
         <quantity>1</quantity>
         <USPrice>148.95</USPrice>
         <comment>Confirm this is electric</comment>
      </item>
      <item partNum="926-AA">
         <productName>Baby Monitor</productName>
         <quantity>1</quantity>
         <USPrice>39.98</USPrice>
         <shipDate>1999-05-21</shipDate>
      </item>
   </items>
</purchaseOrder>
'''
        po1 = load_xml_string(xml)
        po2 = load_xml_string(po1.dump_xml())

        for (path1, obj1), (path2, obj2) in zip(walk(po1), walk(po2)):
            assert path1 == path2
            assert path_to_str(path1) == path_to_str(path2)
            assert type(obj1) is type(obj2)
            if not isinstance(obj1, Object):
                assert obj1 == obj2

        for _ in zip_walk(po1):
            pass

        self.assertEqual(po1.dump(), po2.dump())

    def testDumpLoad(self):

        from tempfile import NamedTemporaryFile as NTF

        class A(Object):
            xmltagname = 'a'
            p = Int.T()

        a1 = A(p=33)
        an = [a1, a1, a1]

        def check1(a, b):
            self.assertEqual(a.p, b.p)

        def checkn(a, b):
            for ea, eb in zip(a, b):
                self.assertEqual(ea.p, eb.p)

        for (a, xdump, xload, check) in [
                (a1, dump, load, check1),
                (a1, dump_xml, load_xml, check1),
                (an, dump_all, load_all, checkn),
                (an, dump_all, iload_all, checkn),
                (an, dump_all_xml, load_all_xml, checkn),
                (an, dump_all_xml, iload_all_xml, checkn)]:

            for header in (False, True, 'custom header'):
                # via string
                s = xdump(a, header=header)
                b = xload(string=s)
                check(a, b)

                # via file
                f = NTF(mode='w+')
                xdump(a, filename=f.name, header=header)
                b = xload(filename=f.name)
                check(a, b)
                f.close()

                # via stream
                for mode in ['w+b', 'w+']:
                    f = NTF(mode=mode)
                    xdump(a, stream=f, header=header)
                    f.seek(0)
                    b = xload(stream=f)
                    check(a, b)
                    f.close()

        b1 = A.load(string=a1.dump())
        check1(a1, b1)

        f = NTF(mode='w+')
        a1.dump(filename=f.name)
        b1 = A.load(filename=f.name)
        check1(a1, b1)
        f.close()

        f = NTF(mode='w+')
        a1.dump(stream=f)
        f.seek(0)
        b1 = A.load(stream=f)
        check1(a1, b1)
        f.close()

        b1 = A.load_xml(string=a1.dump_xml())
        check1(a1, b1)

        f = NTF(mode='w+')
        a1.dump_xml(filename=f.name)
        b1 = A.load_xml(filename=f.name)
        check1(a1, b1)
        f.close()

        f = NTF(mode='w+')
        a1.dump_xml(stream=f)
        f.seek(0)
        b1 = A.load_xml(stream=f)
        check1(a1, b1)
        f.close()

    def testCustomValidator(self):

        class RangeFloat(Float):
            class __T(Float.T):
                def __init__(self, min=0., max=1., *args, **kwargs):
                    Float.T.__init__(self, *args, **kwargs)
                    self.min = min
                    self.max = max

                def validate_extra(self, val):
                    if val < self.min or self.max < val:
                        raise ValidationError('out of range [%g, %g]: %g' % (
                            self.min, self.max, val))

        class A(Object):
            x = RangeFloat.T(min=-1., max=+1., default=0.)

        class B(A):
            y = RangeFloat.T(min=-2., max=+2., default=1.)

            class __T(A.T):
                def validate_extra(self, val):
                    if val.y <= val.x:
                        raise ValidationError('y must be greater than x')

        class C(B):
            class __T(B.T):
                def validate_extra(self, val):
                    if val.y >= val.x:
                        raise ValidationError('x must be greater than y')

        a1 = A()
        a1.validate()

        a2 = A(x=10.)
        with self.assertRaises(ValidationError):
            a2.validate()

        b1 = B()
        b1.validate()

        b2 = B(x=1., y=0.)
        with self.assertRaises(ValidationError):
            b2.validate()

        c1 = C()
        with self.assertRaises(ValidationError):
            c1.validate()

        c2 = C(x=1., y=0.)
        c2.validate()

    def testTypedListClass(self):

        FloatList = make_typed_list_class(Float)

        class A(Object):
            vals = FloatList.T()

        a = A(vals=[1., 2., 3.])
        a.validate()
        a2 = load_string(a.dump())
        assert a2.vals == a.vals

    def testClone(self):

        class A(Object):
            a = Int.T(optional=True)

        class B(Object):
            a_list = List.T(A.T())
            a_tuple = Tuple.T(3, A.T())
            a_dict = Dict.T(Int.T(), A.T())
            b = Float.T()

        a1 = A()
        a2 = A(a=1)
        b = B(
            a_list=[a1, a2],
            a_tuple=(a1, a2, a2),
            a_dict={1: a1, 2: a2},
            b=1.0)
        b_clone = clone(b)
        b.validate()
        b_clone.validate()
        self.assertEqual(b.dump(), b_clone.dump())
        assert b is not b_clone
        assert b.a_list is not b_clone.a_list
        assert b.a_tuple is not b_clone.a_tuple
        assert b.a_list[0] is not b_clone.a_list[0]
        assert b.a_tuple[0] is not b_clone.a_tuple[0]

    def testYPath(self):

        class A(Object):
            i = Int.T()
            i_list = List.T(Int.T())

        class B(Object):
            a_list = List.T(A.T())

        class C(Object):
            b_list = List.T(B.T())

        c = C(b_list=[B(a_list=[A(i=1), A(i=2)]), B(a_list=[A(i=3), A(i=4)])])

        assert get_elements(c, 'b_list[:].a_list[:].i') == [1, 2, 3, 4]
        assert get_elements(c, 'b_list[0].a_list[:].i') == [1, 2]
        assert get_elements(c, 'b_list[:1].a_list[:].i') == [1, 2]
        assert get_elements(c, 'b_list[:1].a_list[1:].i') == [2]
        assert [x.i for x in get_elements(c, 'b_list[-1].a_list[:]')] == [3, 4]

        with self.assertRaises(YPathError):
            get_elements(c, 'b_list[:].x_list[:].i')

        with self.assertRaises(YPathError):
            get_elements(c, 'b_list[:].a_list[:].x')

        with self.assertRaises(YPathError):
            get_elements(c, 'b_list[:].a_list[5]')

        with self.assertRaises(YPathError):
            get_elements(c, 'b_list[:].a_list[:].i.xxx')

        with self.assertRaises(YPathError):
            get_elements(c, '...')

        set_elements(c, 'b_list[1].a_list', [A(i=5), A(i=6)])
        assert get_elements(c, 'b_list[1].a_list[:].i') == [5, 6]

        set_elements(c, 'b_list[:].a_list[:].i', 7)
        assert get_elements(c, 'b_list[:].a_list[:].i') == [7, 7, 7, 7]

        with self.assertRaises(ValidationError):
            set_elements(c, 'b_list[:].a_list[:].i', '7', validate=True)

        set_elements(c, 'b_list[:].a_list[:].i', '7', regularize=True)
        assert get_elements(c, 'b_list[:].a_list[:].i') == [7, 7, 7, 7]

        set_elements(c, 'b_list[:].a_list[:].i_list', [1, 2, 3, 4])
        set_elements(c, 'b_list[:].a_list[:].i_list[1]', 5)
        set_elements(c, 'b_list[:].a_list[:].i_list[:]', 5)
        assert get_elements(c, 'b_list[:].a_list[:].i_list[:]') == [5] * 16

    def testXMLNamespaces(self):
        ns1 = 'http://www.w3.org/TR/html4/'
        ns2 = 'https://www.w3schools.com/furniture'

        class Row(Object):
            xmlns = ns1
            cells = List.T(String.T(xmltagname='td'))

        class Table1(Object):
            xmlns = ns1
            rows = List.T(Row.T(xmltagname='tr'))

        class Table2(Object):
            xmlns = ns2
            name = String.T()
            width = Int.T()
            length = Int.T()

        class Root(Object):
            xmltagname = 'root'
            tables1 = List.T(Table1.T(xmltagname='table'))
            tables2 = List.T(Table2.T(xmltagname='table'))

        doc = '''
<root xmlns:h="http://www.w3.org/TR/html4/"
xmlns:f="https://www.w3schools.com/furniture">

<h:table>
  <h:tr>
    <h:td>Apples</h:td>
    <h:td>Bananas</h:td>
  </h:tr>
</h:table>

<f:table>
  <f:name>African Coffee Table</f:name>
  <f:width>80</f:width>
  <f:length>120</f:length>
</f:table>

</root>
        '''

        o = load_xml(string=doc)
        s = o.dump_xml()
        o2 = load_xml(string=s)
        del o2

        assert isinstance(o.tables1[0], Table1)
        assert isinstance(o.tables2[0], Table2)

    def testXMLNamespaces2(self):
        ns1 = 'http://www.w3.org/TR/html4/'

        class Row(Object):
            xmlns = ns1
            cells = List.T(String.T(xmltagname='td'))

        class Table1(Object):
            xmlns = ns1
            rows = List.T(Row.T(xmltagname='tr'))

        class Root(Object):
            xmltagname = 'root'
            tables1 = List.T(Table1.T(xmltagname='table'))

        doc = '''
<root xmlns:h="http://www.w3.org/TR/html4/">

<h:table>
  <h:tr>
    <h:td>Apples</h:td>
    <h:td>Bananas</h:td>
  </h:tr>
</h:table>

</root>
        '''

        o = load_xml(string=doc)
        assert isinstance(o.tables1[0], Table1)
        s = o.dump_xml()

        o2 = load_xml(string=s)
        assert isinstance(o2.tables1[0], Table1)
        s2 = o2.dump_xml(ns_ignore=True)

        o3 = load_xml(string=s2, ns_hints=['', 'http://www.w3.org/TR/html4/'])
        assert isinstance(o3.tables1[0], Table1)

    def testStyles(self):
        class A(Object):
            s = String.T(yamlstyle=None)
            s_singlequoted = String.T(yamlstyle="'")
            s_doublequoted = String.T(yamlstyle='"')
            s_literal = String.T(yamlstyle='|')
            s_folded = String.T(yamlstyle='>')
            l_block = List.T(String.T(yamlstyle="'"), yamlstyle='block')
            l_flow = List.T(String.T(yamlstyle="'"), yamlstyle='flow')

        a = A(
            s='hello',
            s_singlequoted='hello',
            s_doublequoted='hello',
            s_literal='hello\nhello\n',
            s_folded='hello',
            l_block=['a', 'b', 'c'],
            l_flow=['a', 'b', 'c'])

        a.validate()

        a2 = load_string(a.dump())
        s2 = ('\n'.join(a2.dump().splitlines()[1:]))

        assert s2 == '''
s: hello
s_singlequoted: 'hello'
s_doublequoted: "hello"
s_literal: |
  hello
  hello
s_folded: >-
  hello
l_block:
- 'a'
- 'b'
- 'c'
l_flow: ['a', 'b', 'c']
'''.strip()

    def testEmpty(self):

        class A(Object):
            pass

        class B(Object):
            a = A.T()

        b = B(a=A())
        b2 = load_string(b.dump())
        assert(isinstance(b2.a, A))

    def testNumpyFloat(self):

        class A(Object):
            f = Float.T()
            i = Int.T()

        try:
            import numpy as num

            a = A(f=num.float64(1.0), i=num.int64(1))
            a2 = load_string(a.dump())
            assert a2.f == 1.0
            assert a2.i == 1

            with self.assertRaises(ValidationError):
                a.validate()

            a = A(f=num.float32(1.0), i=num.int32(1))
            a2 = load_string(a.dump())
            assert a2.f == 1.0
            assert a2.i == 1

            with self.assertRaises(ValidationError):
                a.validate()

        except ImportError:
            pass


def makeBasicTypeTest(Type, sample, sample_in=None, xml=False):

    if sample_in is None:
        sample_in = sample

    def basicTypeTest(self):

        class X(Object):
            a = Type.T()
            b = Type.T(optional=True)
            c = Type.T(default=sample)
            d = List.T(Type.T())
            e = Tuple.T(1, Type.T())
            xmltagname = 'x'

        x = X(a=sample_in, e=(sample_in,))
        x.d.append(sample_in)
        if sample_in is not sample:
            with self.assertRaises(ValidationError):
                x.validate()

        x.validate(regularize=sample_in is not sample)
        self.assertEqualNanAware(sample, x.a)

        if not xml:
            x2 = load_string(x.dump())

        else:
            x2 = load_xml_string(x.dump_xml())

        self.assertEqualNanAware(x.a, x2.a)
        self.assertIsNone(x.b)
        self.assertIsNone(x2.b)
        self.assertEqualNanAware(sample, x.c)
        self.assertEqualNanAware(sample, x2.c)
        self.assertTrue(isinstance(x.d, list))
        self.assertTrue(isinstance(x2.d, list))
        self.assertEqualNanAware(x.d[0], sample)
        self.assertEqualNanAware(x2.d[0], sample)
        self.assertEqual(len(x.d), 1)
        self.assertEqual(len(x2.d), 1)
        self.assertTrue(isinstance(x.e, tuple))
        self.assertTrue(isinstance(x2.e, tuple))
        self.assertEqualNanAware(x.e[0], sample)
        self.assertEqualNanAware(x2.e[0], sample)
        self.assertEqual(len(x.e), 1)
        self.assertEqual(len(x2.e), 1)

    return basicTypeTest


for Type in samples:
    for isample, sample in enumerate(samples[Type]):
        for xml in (False, True):
            name = 'testBasicType' + Type.__name__ + str(isample) \
                + ['', 'XML'][xml]
            func = makeBasicTypeTest(Type, sample, xml=xml)
            func.__name__ = name
            setattr(GutsTestCase, name, func)
            del func

for Type in regularize:
    for isample, (sample_in, sample) in enumerate(regularize[Type]):
        for xml in (False, True):
            name = 'testBasicTypeRegularize' + Type.__name__ + str(isample) \
                + ['', 'XML'][xml]
            func = makeBasicTypeTest(
                Type, sample, sample_in=sample_in, xml=xml)
            func.__name__ = name
            setattr(GutsTestCase, name, func)
            del func


if __name__ == '__main__':
    unittest.main()



# python 2/3
from __future__ import division, print_function, absolute_import

from builtins import range
import time
import unittest
from collections import defaultdict

import numpy as num

from . import common
from pyrocko import util, io, datacube_ext
from pyrocko.io import datacube


class DataCubeTestCase(unittest.TestCase):

    def test_load(self):
        fpath = common.test_data_file('test1.cube')

        traces_h = io.load(fpath, getdata=False, format='detect')
        traces_d = io.load(fpath, getdata=True, format='detect')

        mimas = [
            (30572, 87358),
            (46168, 80639),
            (53107, 73119)]

        for tr_h, tr_d, (mi, ma) in zip(traces_h, traces_d, mimas):
            assert tr_h.tmin == tr_d.tmin
            assert tr_h.tmax == tr_d.tmax
            assert tr_d.ydata.min() == mi
            assert tr_d.ydata.max() == ma

    def test_load_partial(self):
        fpath = common.test_data_file('test2.cube')
        f = open(fpath, 'r')
        header, da1, gps_tags, nsamples, bookmarks = datacube_ext.load(
            f.fileno(), 2, 0, -1, None)

        for ioff in (0, 10, 1040000, 1048576, 2000000, 1000000):
            f.seek(0)
            header, da2, gps_tags, nsamples, _ = datacube_ext.load(
                f.fileno(), 2, ioff, 10, None)

            f.seek(0)
            header, da3, gps_tags, nsamples, _ = datacube_ext.load(
                f.fileno(), 2, ioff, 10, bookmarks)

            for a1, a2, a3 in zip(da1, da2, da3):
                assert num.all(a1[ioff:ioff+10] == a2) and num.all(a2 == a3)

        f.close()

    def test_interpolate_or_not(self):
        fpath = common.test_data_file('test2.cube')
        trs = {}
        for imode in ('off', 'sinc'):
            trs[imode] = list(datacube.iload(fpath, interpolation=imode))[:1]

            for tr in trs[imode]:
                tr.set_codes(location='i=%s' % imode)

        # import pylab as lab
        for cha in ['p0']:  # 'p1', 'p2']:
            t1 = [tr for tr in trs['off'] if tr.channel == cha][0]
            t2 = [tr for tr in trs['sinc'] if tr.channel == cha][0]
            it = 0
            nt = min(t1.ydata.size, t2.ydata.size)

            dd = []
            nb = int(600. / t1.deltat)
            while it < nt:
                y1 = t1.ydata[it:it+nb]
                y2 = t2.ydata[it:it+nb]
                dd.append(abs(num.mean(y1) - num.mean(y2)))
                assert dd[-1] < 1.0
                it += nb

            # t = num.arange(len(dd))*600.
            # d = num.array(dd)
            # lab.plot(t / 3600., d)

        # lab.show()

        # trace.snuffle(trs['off'] + trs['sinc'])

    def test_timing_context(self):
        fpath = common.test_data_file('test2.cube')
        datacube.get_extended_timing_context(fpath)

    def benchmark_load(self):
        mode = {
            0: 'get time range',
            1: 'get gps only',
            2: 'get samples'}

        fpath = common.test_data_file('test2.cube')
        for irep in range(2):
            for loadflag in (0, 1, 2):
                f = open(fpath, 'r')
                t0 = time.time()
                header, data_arrays, gps_tags, nsamples, bookmarks = \
                    datacube_ext.load(f.fileno(), loadflag, 0, -1, None)

                f.close()
                t1 = time.time()
                print('%s: %10.3f' % (mode[loadflag], t1 - t0))

            t0 = time.time()
            trs = io.load(fpath, format='datacube')
            t1 = time.time()
            print('with interpolation: %10.3f' % (t1 - t0))
            del trs

    def test_leapsecond(self):
        fns = map(common.test_data_file, ['leapsecond_dec.cube',
                                          'leapsecond_jan.cube'])

        trs = defaultdict(list)
        for fn in fns:
            for tr in datacube.iload(fn):
                trs[tr.channel].append(tr)

        for cha in trs.keys():
            tra, trb, trc = trs[cha]
            assert abs(
                tra.tmax - (util.stt('2017-01-01 00:00:01') - tra.deltat)) \
                < tra.deltat * 0.001
            assert abs(
                trb.tmin - util.stt('2017-01-01 00:00:00')) \
                < trb.deltat * 0.001


if __name__ == "__main__":
    util.setup_logging('test_io', 'warning')
    unittest.main()



from builtins import str
from pyrocko.gui.snuffling import Snuffling, Param, Switch, Choice
from pyrocko.gui.marker import PhaseMarker
from pyrocko import orthodrome
from pyrocko import gf
from pyrocko import cake
import numpy as num


class CakePhase(Snuffling):
    '''
    <html>
    <head>
    <style type="text/css">
        body { margin-left:10px };
    </style>
    </head>
    <body>
    <h1 align="center">Theoretical Phase Arrivals</h1>
    <p>
    This snuffling uses pyrocko's
    <a href="http://emolch.github.io/pyrocko/v0.3/cake_doc.html">Cake</a>
    module to calculate seismic rays for layered earth models. </p>
    <p>
    <b>Parameters:</b><br />
        <b>&middot; Global shift</b>  -  Add time onset to phases. <br />
        <b>&middot; Add Model</b>  -  Add a model to drop down menu. <br />
        <b>&middot; Add Phase</b>  -  Add a phase definition.
            (GUI reset required)<br />
    </p>
    <p>
    Instructions and information on Cake's syntax of seismic rays can be
    found in the
    <a href="http://emolch.github.io/pyrocko/
    v0.3/cake_doc.html#cmdoption-cake--phase">Cake documentation</a>.
    </p>
    </body>
    </html>
    '''

    def setup(self):
        self.set_name('Cake Phase')

        # self._phase_names = ('PmP ~S ~P ~P(moho)s ~P(sill-top)s'
        #                       ' ~P(sill-bottom)s Pdiff').split()
        self._phase_names = ('~P Pg Sg pP p P Pdiff PKP PcP PcS PKIKP pPKIKP'
                             ' SSP PPS SPP PSP SP PS ~PS ~SP Pn s S Sn PP PPP'
                             ' ScS Sdiff SS SSS PcP SKS SKIKS').split()

        for iphase, name in enumerate(self._phase_names):
            self.add_parameter(Switch(name, 'wantphase_%i' % iphase,
                                      iphase == 0))

        model_names = cake.builtin_models()
        model_names = [
                'Cake builtin: %s' % model_name for model_name in model_names]

        self._engine = gf.LocalEngine(use_config=True)
        store_ids = self._engine.get_store_ids()

        for store_id in store_ids:
            model_names.append('GF Store: %s' % store_id)

        self._models = model_names

        self.model_choice = Choice('Model', 'chosen_model',
                                   'ak135-f-continental.m', self._models)

        self.add_parameter(self.model_choice)

        self.add_parameter(Param('Global shift', 'tshift', 0., -20., 20.))
        self.add_parameter(Switch('Use station depth',
                                  'use_station_depth', False))
        self.add_trigger('Add Phase', self.add_phase_definition)
        self.add_trigger('Add Model', self.add_model_to_choice)
        self.add_trigger('Plot Model', self.plot_model)
        self.add_trigger('Plot Rays', self.plot_rays)

        self._phases = {}
        self._model = None

    def panel_visibility_changed(self, bool):
        pass

    def wanted_phases(self):
        try:
            wanted = []
            for iphase, name in enumerate(self._phase_names):
                if getattr(self, 'wantphase_%i' % iphase):
                    if name in self._phases:
                        phases = self._phases[name]
                    else:
                        if name.startswith('~'):
                            phases = [cake.PhaseDef(name[1:])]
                        else:
                            phases = cake.PhaseDef.classic(name)

                        self._phases[name] = phases
                        for pha in phases:
                            pha.name = name

                    wanted.extend(phases)
        except (cake.UnknownClassicPhase, cake.PhaseDefParseError) as e:
            self.fail(str(e))

        return wanted

    def call(self, plot_rays=False):

        self.cleanup()
        wanted = self.wanted_phases()

        if not wanted:
            return

        event, stations = self.get_active_event_and_stations()

        if not stations:
            self.fail('No station information available.')

        self.update_model()
        model = self._model[1]

        depth = event.depth
        if depth is None:
            depth = 0.0

        allrays = []
        alldists = []
        for station in stations:
            dist = orthodrome.distance_accurate50m(event, station)
            alldists.append(dist)

            if self.use_station_depth:
                rdepth = station.depth
            else:
                rdepth = 0.0

            multi_dists = []
            nmax = 1
            for i in range(0, nmax):
                multi_dists.append(dist*cake.m2d + 360.*i)
                multi_dists.append((i+1)*360. - dist*cake.m2d)

            rays = model.arrivals(
                phases=wanted,
                distances=multi_dists,
                zstart=depth,
                zstop=rdepth)

            for ray in rays:
                time = ray.t
                name = ray.given_phase().name
                incidence_angle = ray.incidence_angle()
                takeoff_angle = ray.takeoff_angle()

                time += event.time + self.tshift
                m = PhaseMarker([(station.network, station.station, '*', '*')],
                                time, time, 2,
                                phasename=name,
                                event=event,
                                incidence_angle=incidence_angle,
                                takeoff_angle=takeoff_angle)
                self.add_marker(m)

            allrays.extend(rays)

        if plot_rays:
            fig = self.figure(name='Ray Paths')
            from pyrocko import cake_plot
            cake_plot.my_rays_plot(model, None, allrays, depth, 0.0,
                                   num.array(alldists)*cake.m2d,
                                   axes=fig.gca())

            fig.canvas.draw()

    def update_model(self):
        if not self._model or self._model[0] != self.chosen_model:
            if self.chosen_model.startswith('Cake builtin: '):
                load_model = cake.load_model(
                    self.chosen_model.split(': ', 1)[1])

            elif self.chosen_model.startswith('GF Store: '):
                store_id = self.chosen_model.split(': ', 1)[1]

                load_model = self._engine.get_store(store_id)\
                    .config.earthmodel_1d
            else:
                load_model = cake.load_model(self.chosen_model)

            self._model = (self.chosen_model, load_model)

    def update_model_choices(self):
        self.set_parameter_choices('chosen_model', self._models)

    def add_model_to_choice(self):
        '''Called from trigger 'Add Model'.

        Adds another choice to the drop down 'Model' menu.
        '''

        in_model = self.input_filename('Load Model')
        if in_model not in self._models:
            self._models.append(in_model)
            self.update_model_choices()

        self.set_parameter('chosen_model', in_model)
        self.call()

    def add_phase_definition(self):
        ''' Called from trigger 'Add Phase Definition'.

        Adds another phase option.
        Requires a reset of the GUI.
        '''
        phase_def = str(self.input_dialog('Add New Phase',
                                          'Enter Phase Definition'))
        self._phase_names.append(phase_def)

        self.add_parameter(
            Switch(phase_def,
                   'wantphase_%s' % str(len(self._phase_names)-1), True))
        self.reset_gui(reloaded=True)
        self.call()

    def plot_model(self):
        self.update_model()

        from pyrocko import cake_plot

        fig = self.figure(name='Model: %s' % self._model[0])

        cake_plot.my_model_plot(self._model[1], axes=fig.gca())

        fig.canvas.draw()

    def plot_rays(self):
        self.call(plot_rays=True)


def __snufflings__():
    '''Returns a list of snufflings to be exported by this module.'''

    return [CakePhase()]



from __future__ import division, print_function, absolute_import
from builtins import range

import unittest
import numpy as num
from numpy.testing import assert_array_less

from pyrocko.dataset import gshhg
from pyrocko import util

plot = False


class BB(object):
    west = 5.
    east = 20.
    south = 35.
    north = 55.

    tpl = (west, east, south, north)


class BBLakes(object):
    west = 4.41
    east = 13.14
    south = 42.15
    north = 49.32

    tpl = (west, east, south, north)


class BBPonds(object):
    west, east, south, north = (
        304.229444, 306.335556, -3.228889, -1.1358329999999999)

    tpl = (west, east, south, north)


class BBOtherSide(object):
    west = -5.
    east = 5.
    south = 30.
    north = 50.

    tpl = (west, east, south, north)


class GSHHGTest(unittest.TestCase):
    def setUp(self):
        self.gshhg = gshhg.GSHHG.intermediate()

    def test_polygon_loading_points(self):
        for ipoly in range(10):
            self.gshhg.polygons[ipoly].points

    def test_polygon_contains_point(self):
        poly = self.gshhg.polygons[-1]
        p_within = num.array(
            [poly.south + (poly.north - poly.south) / 2,
             poly.west + (poly.east - poly.west) / 2])

        p_outside = num.array(
            [poly.north + (poly.north - poly.south) / 2,
             poly.east + (poly.east - poly.west) / 2])

        assert poly.contains_point(p_within) is True
        assert poly.contains_point(p_outside) is False

    def test_latlon(self):
        p = self.gshhg.polygons[0]

        assert_array_less(num.zeros_like(p.lons) - 0.001, p.lons)
        assert_array_less(p.lats, num.ones_like(p.lats) * 90 + 0.001)

    def test_polygon_level_selection(self):
        for p in self.gshhg.polygons:
            p.is_land()
            p.is_island_in_lake()
            p.is_pond_in_island_in_lake()
            p.is_antarctic_icefront()
            p.is_antarctic_grounding_line()

    def test_polygon_contains_points(self):
        poly = self.gshhg.polygons[0]

        points = num.array(
                [num.random.uniform(BB.south, BB.north, size=100),
                 num.random.uniform(BB.east, BB.west, size=100)]).T
        pts = poly.contains_points(points)

        if plot:
            import matplotlib.pyplot as plt
            import cartopy.crs as ccrs

            points = points[pts]

            colors = num.ones(points.shape[0]) * pts[pts]
            ax = plt.axes(projection=ccrs.PlateCarree())
            ax.set_global()

            self.plot_polygons([poly], ax)
            ax.scatter(points[:, 1], points[:, 0], c=colors, s=.5,
                       transform=ccrs.Geodetic())
            plt.show()

    def test_bounding_box_select(self):
        p = self.gshhg.get_polygons_within(*BB.tpl)

        if plot:
            import matplotlib.pyplot as plt
            import cartopy.crs as ccrs

            from matplotlib.patches import Rectangle

            ax = plt.axes(projection=ccrs.PlateCarree())
            ax.set_global()
            self.plot_polygons(p, ax)

            ax.add_patch(
                Rectangle([BB.west, BB.south],
                          width=BB.east-BB.west, height=BB.north-BB.south,
                          alpha=.2))
            plt.show()

    def test_mask_land(self):
        poly = self.gshhg.get_polygons_within(*BBPonds.tpl)

        points = num.array(
                [num.random.uniform(BBPonds.south, BBPonds.north, size=1000),
                 num.random.uniform(BBPonds.east, BBPonds.west, size=1000)]).T
        pts = self.gshhg.get_land_mask(points)

        if plot:
            import cartopy.crs as ccrs
            import matplotlib.pyplot as plt

            points = points
            colors = num.ones(points.shape[0]) * pts

            # colors = num.ones(points.shape[0]) * pts
            ax = plt.axes(projection=ccrs.PlateCarree())
            ax.set_global()

            self.plot_polygons(poly, ax)
            ax.scatter(points[:, 1], points[:, 0], c=colors, s=.5,
                       transform=ccrs.Geodetic(), zorder=2)
            plt.show()

    def test_is_point_on_land(self):

        point = (46.455289, 6.494283)
        p = self.gshhg.get_polygons_at(*point)
        assert self.gshhg.is_point_on_land(*point) is False

        if plot:
            import matplotlib.pyplot as plt
            import cartopy.crs as ccrs

            ax = plt.axes(projection=ccrs.PlateCarree())
            ax.set_global()
            self.plot_polygons(p, ax)
            ax.scatter(point[1], point[0], transform=ccrs.Geodetic())

            plt.show()

    @staticmethod
    def plot_polygons(polygons, ax, **kwargs):
        from matplotlib.patches import Polygon
        import numpy as num
        import cartopy.crs as ccrs
        from pyrocko.plot import mpl_color

        args = {
            'edgecolor': 'red',
        }
        args.update(kwargs)

        colormap = [
            mpl_color('aluminium2'),
            mpl_color('skyblue1'),
            mpl_color('aluminium4'),
            mpl_color('skyblue2'),
            mpl_color('white'),
            mpl_color('aluminium1')]

        map(ax.add_patch, [Polygon(num.fliplr(p.points),
                                   transform=ccrs.Geodetic(),
                                   facecolor=colormap[p.level_no-1],
                                   **args)
                           for p in polygons])

    def test_pond(self):
        for poly in self.gshhg.polygons:
            if poly.is_pond_in_island_in_lake():
                (w, e, s, n) = poly.get_bounding_box()
                # print (w-1., e+1, s-1., n+1)
                polys2 = self.gshhg.get_polygons_within(w-1., e+1, s-1., n+1)

                if plot:
                    import matplotlib.pyplot as plt
                    import cartopy.crs as ccrs

                    ax = plt.axes(projection=ccrs.PlateCarree())
                    ax.set_global()
                    self.plot_polygons(polys2, ax)

                    plt.show()

    def test_other_side(self):
        bb = BBOtherSide
        poly = self.gshhg.get_polygons_within(*bb.tpl)

        points = num.array(
                [num.random.uniform(bb.south, bb.north, size=1000),
                 num.random.uniform(bb.east, bb.west, size=1000)]).T
        pts = self.gshhg.get_land_mask(points)

        if plot:
            import cartopy.crs as ccrs
            import matplotlib.pyplot as plt

            points = points
            colors = num.ones(points.shape[0]) * pts

            # colors = num.ones(points.shape[0]) * pts
            ax = plt.axes(projection=ccrs.PlateCarree())
            ax.set_global()

            self.plot_polygons(poly, ax)
            ax.scatter(points[:, 1], points[:, 0], c=colors, s=.5,
                       transform=ccrs.Geodetic(), zorder=2)
            plt.show()


if __name__ == "__main__":
    plot = False
    util.setup_logging('test_gshhg', 'debug')
    unittest.main(exit=False)



from __future__ import division, print_function, absolute_import
import os
import math
import tempfile
import shutil
import unittest
import numpy as num
from numpy.testing import assert_allclose
from matplotlib import image, pyplot as plt

from pyrocko import util
from pyrocko.plot import gmtpy
from pyrocko.plot.gmtpy import cm, inch, golden_ratio

from . import common

plot = False


@unittest.skipUnless(
    gmtpy.have_gmt(), 'GMT not available')
@unittest.skipUnless(
    gmtpy.have_pixmaptools(), '`pdftocairo` or `convert` not available')
class GmtPyTestCase(unittest.TestCase):

    def setUp(self):
        self.tempdir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def fpath(self, fn):
        return os.path.join(self.tempdir, fn)

    def fpath_ref(self, fn):
        try:
            return common.test_data_file(fn)
        except util.DownloadError:
            return common.test_data_file_no_download(fn)

    def compare_with_ref(self, fname, tolerance, show=False):
        fpath = self.fpath(fname)
        fpath_ref = self.fpath_ref(fname)

        if not os.path.exists(fpath_ref):
            shutil.copy(fpath, fpath_ref)

        img = image.imread(fpath)
        img_ref = image.imread(fpath_ref)
        self.assertEqual(img.shape, img_ref.shape)
        d = num.abs(img - img_ref)
        merr = num.mean(d)
        if (merr > tolerance or show) and plot:
            fig = plt.figure()
            axes1 = fig.add_subplot(1, 3, 1, aspect=1.)
            axes2 = fig.add_subplot(1, 3, 2, aspect=1.)
            axes3 = fig.add_subplot(1, 3, 3, aspect=1.)
            axes1.imshow(img)
            axes1.set_title('Candidate')
            axes2.imshow(img_ref)
            axes2.set_title('Reference')
            axes3.imshow(d)
            axes3.set_title('Mean abs difference: %g' % merr)
            plt.show()
            plt.close(fig)

        assert merr <= tolerance

    def test_basic(self):
        for version in gmtpy.all_installed_gmt_versions():
            width = 8.0 * inch
            height = 9.0 * inch
            resolution = 72

            gmt = gmtpy.GMT(version=version, config_papersize=(width, height))

            gmt.pscoast(
                X=0,
                Y=0,
                R='g',
                J='E32/30/170/8i',
                B='10g10',
                D='c',
                A=10000,
                S=(114, 159, 207),
                G=(233, 185, 110),
                W='thinnest')

            gmt.dump('test')
            gmt.load('test')

            for oversample in (1, 2):
                fname = 'gmtpy_test_basic_o%i.png' % oversample
                fpath = self.fpath(fname)
                gmt.save(fpath, resolution=resolution, oversample=oversample)

                self.compare_with_ref(fname, 0.03)

                img = image.imread(fpath, format='png')
                self.assertEqual(img.shape, (
                    int(round(resolution*height/inch)),
                    int(round(resolution*width/inch)), 3))

    def test_basic2(self):
        for version in gmtpy.all_installed_gmt_versions():
            if version.startswith('5'):
                gmt = gmtpy.GMT(
                    version=version,
                    config={'MAP_FRAME_TYPE': 'fancy'},
                    eps_mode=True)
            else:
                gmt = gmtpy.GMT(
                    version=version,
                    config={'BASEMAP_TYPE': 'fancy'})

            layout = gmt.default_layout()
            widget = layout.get_widget()

            xax = gmtpy.Ax(label='Lon', mode='min-max')
            yax = gmtpy.Ax(label='Lat', mode='min-max')
            scaler = gmtpy.ScaleGuru([([5, 15], [52, 58])], axes=(xax, yax))

            par = scaler.get_params()
            lon0 = (par['xmin'] + par['xmax'])/2.
            lat0 = (par['ymin'] + par['ymax'])/2.
            sll = '%g/%g' % (lon0, lat0)
            widget['J'] = '-JM' + sll + '/%(width)gp'

            widget['J'] = '-JM' + sll + '/%(width)gp'
            scaler['B'] = \
                '-B%(xinc)gg%(xinc)g:%(xlabel)s:' \
                '/%(yinc)gg%(yinc)g:%(ylabel)s:WSen'

            aspect = gmtpy.aspect_for_projection(
                version, *(widget.J() + scaler.R()))

            aspect = 1.045
            widget.set_aspect(aspect)

            gmt.pscoast(D='h', W='1p,red', *(widget.JXY() + scaler.R()))
            gmt.psbasemap(*(widget.JXY() + scaler.BR()))

            fname = 'gmtpy_test_basic2.png'
            fpath = self.fpath(fname)
            gmt.save(fpath, resolution=75, bbox=layout.bbox())

            self.compare_with_ref(fname, 0.01, show=False)

    def test_layout(self):
        x = num.linspace(0., math.pi*6, 1001)
        y1 = num.sin(x) * 1e-9
        y2 = 2.0 * num.cos(x) * 1e-9

        xax = gmtpy.Ax(label='Time', unit='s')
        yax = gmtpy.Ax(
            label='Amplitude', unit='m', scaled_unit='nm',
            scaled_unit_factor=1e9, approx_ticks=5, space=0.05)

        guru = gmtpy.ScaleGuru([(x, y1), (x, y2)], axes=(xax, yax))

        for version in gmtpy.all_installed_gmt_versions():
            width = 8*inch
            height = 3*inch

            gmt = gmtpy.GMT(
                version=version,
                config_papersize=(width, height))

            layout = gmt.default_layout()
            widget = layout.get_widget()

            gmt.draw_layout(layout)

            gmt.psbasemap(*(widget.JXY() + guru.RB(ax_projection=True)))
            gmt.psxy(
                in_columns=(x, y1), W='1p,red', *(widget.JXY() + guru.R()))
            gmt.psxy(
                in_columns=(x, y2), W='1p,blue', *(widget.JXY() + guru.R()))

            fname = 'gmtpy_test_layout.png'
            fpath = self.fpath(fname)
            gmt.save(fpath)

            self.compare_with_ref(fname, 0.01)

    def test_grid_layout(self):
        for version in gmtpy.all_installed_gmt_versions():
            gmt = gmtpy.GMT(version=version, config_papersize='a3')
            nx, ny = 2, 5
            grid = gmtpy.GridLayout(nx, ny)

            layout = gmt.default_layout()
            layout.set_widget('center', grid)

            widgets = []
            for iy in range(ny):
                for ix in range(nx):
                    inner = gmtpy.FrameLayout()
                    inner.set_fixed_margins(
                        1.*cm*golden_ratio, 1.*cm*golden_ratio, 1.*cm, 1.*cm)

                    grid.set_widget(ix, iy, inner)
                    inner.set_vertical(0, (iy+1.))
                    widgets.append(inner.get_widget('center'))

            gmt.draw_layout(layout)
            for widget in widgets:
                x = num.linspace(0., 10., 5)
                y = num.sin(x)
                xax = gmtpy.Ax(approx_ticks=4, snap=True)
                yax = gmtpy.Ax(approx_ticks=4, snap=True)
                guru = gmtpy.ScaleGuru([(x, y)], axes=(xax, yax))
                gmt.psbasemap(*(widget.JXY() + guru.RB(ax_projection=True)))
                gmt.psxy(in_columns=(x, y), *(widget.JXY() + guru.R()))

            fname = 'gmtpy_test_grid_layout.png'
            fpath = self.fpath(fname)
            gmt.save(fpath, resolution=75)

            self.compare_with_ref(fname, 0.01)

    def test_simple(self):

        x = num.linspace(0., 2*math.pi)
        y = num.sin(x)
        y2 = num.cos(x)

        for version in gmtpy.all_installed_gmt_versions():
            for ymode in ['off', 'symmetric', 'min-max', 'min-0', '0-max']:
                plot = gmtpy.Simple(gmtversion=version, ymode=ymode)
                plot.plot((x, y), '-W1p,%s' % gmtpy.color('skyblue2'))
                plot.plot((x, y2), '-W1p,%s' % gmtpy.color(
                    gmtpy.color_tup('scarletred2')))
                plot.text((3., 0.5, 'hello'), size=20.)
                fname = 'gmtpy_test_simple_%s.png' % ymode
                fpath = self.fpath(fname)
                plot.save(fpath)
                self.compare_with_ref(fname, 0.01, show=False)

    def test_simple_density(self):
        x = num.linspace(0., 2.*math.pi, 50)
        y = num.linspace(0., 2.*math.pi, 50)

        x2 = num.tile(x, y.size)
        y2 = num.repeat(y, x.size)
        z2 = num.sin(x2) * num.sin(y2)

        for version in gmtpy.all_installed_gmt_versions():
            for method in ['surface', 'triangulate', 'fillcontour']:
                plot = gmtpy.Simple(gmtversion=version, with_palette=True)
                plot.density_plot((x2, y2, z2), method=method)
                fname = 'gmtpy_test_simple_density_%s.png' % method
                fpath = self.fpath(fname)
                plot.save(fpath)
                self.compare_with_ref(fname, 0.02)

    def test_grid_data(self):
        x = num.linspace(0., 2.*math.pi, 100)
        y = num.linspace(0., 2.*math.pi, 100)

        x2 = num.tile(x, y.size)
        y2 = num.repeat(y, x.size)
        z2 = num.sin(x2) * num.sin(y2)

        xf, yf, zf = gmtpy.griddata_auto(x2, y2, z2)
        assert (xf.size, yf.size, zf.size) == (100, 100, 100*100)
        x3, y3, z3 = gmtpy.tabledata(xf, yf, zf)

        assert_allclose(x3, x2, atol=1e-7)
        assert_allclose(y3, y2, atol=1e-7)
        assert_allclose(z3, z2, atol=1e-7)

        xf2, yf2, zf2 = gmtpy.doublegrid(xf, yf, zf)
        assert (xf2.size, yf2.size, zf2.size) == (199, 199, 199*199)

        fn = self.fpath('grid.nc')
        for naming in ['xy', 'lonlat']:
            gmtpy.savegrd(xf, yf, zf, fn, naming=naming, title='mygrid')
            xf3, yf3, zf3 = gmtpy.loadgrd(fn)
            assert_allclose(xf3, xf)
            assert_allclose(yf3, yf)
            assert_allclose(zf3, zf)

    def test_text_box(self):

        for version in gmtpy.all_installed_gmt_versions():
            s = gmtpy.text_box('Hello', gmtversion=version)
            assert_allclose(s, (25.8, 9.), rtol=0.1)
            s = gmtpy.text_box(
                'Abc def ghi jkl mno pqr stu vwx yz',
                gmtversion=version)
            assert_allclose(s, (179.9, 12.3), rtol=0.01)

    def test_override_args(self):
        x = num.array([0, 0.5, 1, 0])
        y = num.array([0, 1, 0, 0])
        width = 300
        height = 100
        config_papersize = (width, height)

        for version in gmtpy.all_installed_gmt_versions():
            gmt = gmtpy.GMT(
                version=version,
                config_papersize=config_papersize)

            for i, cutoff in enumerate([30, 90]):
                gmt.psxy(
                    in_columns=(i*2+x, y),
                    W='10p,red',
                    J='X%gp/%gp' % (width, height),
                    X=0,
                    Y=0,
                    R=(-1, 4, -1, 2),
                    config={
                        'PS_MITER_LIMIT': '%i' % cutoff})

            fname = 'gmtpy_test_override.png'
            fpath = self.fpath(fname)
            gmt.save(fpath)
            self.compare_with_ref(fname, 0.001, show=False)


if __name__ == "__main__":
    plot = True
    util.setup_logging('test_gmtpy', 'warning')
    unittest.main()




import unittest
import logging
from pyrocko import util
from . import common

logger = logging.getLogger('pyrocko.test.test_urls')

try:
    from urllib2 import (Request, HTTPError, urlopen)
except ImportError:
    from urllib.request import (Request, urlopen)
    from urllib.error import HTTPError


class UrlsTestCase(unittest.TestCase):

    @common.require_internet
    def test_url_alive(self):
        # Test urls which are used as references in pyrocko if they still
        # exist.
        to_check = [
            ('http://nappe.wustl.edu/antelope/css-formats/wfdisc.htm',
             'pyrocko.css'),
            ('http://www.ietf.org/timezones/data/leap-seconds.list',
             'pyrocko.config'),
            ('http://stackoverflow.com/questions/2417794/', 'cake_plot'),
            ('http://igppweb.ucsd.edu/~gabi/rem.html', 'crust2x2_data'),
            ('http://kinherd.org/pyrocko_data/gsc20130501.txt', 'crustdb'),
            ('http://download.geonames.org/export/dump/', 'geonames'),
            ('http://emolch.github.io/gmtpy/', 'gmtpy'),
            ('http://www.apache.org/licenses/LICENSE-2.0', 'kagan.py'),
            ('http://www.opengis.net/kml/2.2', 'model'),
            ('http://maps.google.com/mapfiles/kml/paddle/S.png', 'model'),
            ('http://de.wikipedia.org/wiki/Orthodrome', 'orthodrome'),
            ('http://peterbird.name/oldFTP/PB2002', 'tectonics'),
            ('http://gsrm.unavco.org/model', 'tectonics'),
            ('http://stackoverflow.com/questions/19332902/', 'util'),
        ]

        for url, label in to_check:
            try:
                try:
                    req = Request(url, method='HEAD')
                except TypeError:
                    req = Request(url)

                f = urlopen(req)
                f.close()
            except HTTPError as e:
                logger.warn('%s - %s referenced in pyrocko.%s' %
                            (e, url, label))


if __name__ == '__main__':
    util.setup_logging('test_urls', 'warning')
    unittest.main()
